from typing import Iterable, Tuple

from jinja2.ext import Extension
from jinja2.lexer import Token


class SourceMapper:
    """
    Container class that can be used to map lines of rendered text to the
    template file and line that generated that text.
    """

    def __init__(self, line_breaks: Iterable[Tuple[int, int, str]]) -> None:
        self.line_breaks = tuple(line_breaks)

    def get_source_line_data(self, pos: int) -> Tuple[str, int]:
        """
        Given a character position in the rendered document return the template
        file and line number that produced it.
        """
        if not self.line_breaks:
            return "<none>", -1

        idx = 0
        while idx + 1 < len(self.line_breaks) and pos > self.line_breaks[idx][0]:
            idx += 1
        return (
            self.line_breaks[idx][2],
            self.line_breaks[idx][1],
        )

    def get_source_line(self, pos: int) -> str:
        """
        Like `get_source_line_data` except return a single string in the form
        {file}:{line_no}.
        """
        result = self.get_source_line_data(pos)
        return f"{result[0]}:{result[1]}"


class SourceMapperExtension(Extension):
    """
    Jinja extension used to correlate template files and line numbers to each
    character of output in the rendered document. It does this by inserting
    instructions to print metadata within the template and removing these
    metadata chunks when passing the template output back through the render
    function.
    """

    def filter_stream(self, stream):
        """
        Extension filter function that adds output to track the template file
        and line numbers in the output stream.
        """
        yield Token(1, "block_begin", "{%")
        yield Token(1, "name", "print")
        yield Token(1, "string", f"\u00001;{stream.filename}\u0000")
        yield Token(1, "block_end", "%}")

        for token in stream:
            if token.type == "data":
                yield Token(1, "block_begin", "{%")
                yield Token(1, "name", "print")
                yield Token(
                    1, "string", f"\u0000{token.lineno};{stream.filename}\u0000"
                )
                yield Token(1, "block_end", "%}")
            yield token

    def render(self, generator) -> Tuple[str, SourceMapper]:
        """
        Given a generator created from jinja_env.generate, output the complete
        rendered document and a SourceMapper object to map each character to a
        file and line number in the source template.
        """
        cur_pos = 0
        cur_lineno = 0
        cur_filename = "<none>"
        line_breaks = []
        result = []

        chunks = "".join(generator).split("\u0000")
        for ind, data in enumerate(chunks):
            if ind % 2 == 1:
                s1 = data.find(";")
                cur_lineno = int(data[:s1])
                cur_filename = data[s1 + 1 :]
                line_breaks.append((cur_pos, cur_lineno, cur_filename))
            else:
                result.append(data)
                for idx, ch in enumerate(data):
                    if ch == "\n":
                        line_breaks.append((cur_pos + idx, cur_lineno, cur_filename))
                        cur_lineno += 1
                cur_pos += len(data)

        return "".join(result), SourceMapper(line_breaks)

    def parse(self, parser):
        raise NotImplementedError("No parse method needed")



import argparse

from tplbuild.cmd.utility import CliUtility
from tplbuild.images import SourceImage
from tplbuild.tplbuild import TplBuild


class SourceLookupUtility(CliUtility):
    """CLI utility entrypoint for looking up source images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "print the full image name with digest of source images"
        parser.add_argument(
            "source",
            nargs="*",
            help="Source images to update in repo:tag format",
        )
        parser.add_argument(
            "--platform",
            action="append",
            help="Platforms to lookup. Defautls to the current executor platform.",
        )
        parser.add_argument(
            "--digest-only",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Only print the digest of the source images",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        platform = args.platform or await tplbld.get_default_platform()

        for image in args.source:
            colon = image.find(":")
            if colon == -1:
                repo, tag = image, "latest"
            else:
                repo, tag = image[:colon], image[colon + 1 :]
            resolved_image = await tplbld.resolve_image(
                SourceImage(repo=repo, tag=tag, platform=platform),
                check_only=True,
            )
            if args.digest_only:
                print(resolved_image.digest)
            else:
                print(f"{repo}@{resolved_image.digest}")

        return 0



import argparse

from aioregistry import parse_image_name

from tplbuild.cmd.utility import CliUtility
from tplbuild.exceptions import TplBuildException, TplBuildNoSourceImageException
from tplbuild.graph import visit_graph
from tplbuild.images import ImageDefinition, SourceImage
from tplbuild.tplbuild import TplBuild


class SourceUpdateUtility(CliUtility):
    """CLI utility entrypoint for updating source images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = (
            "explicitly update source image digests from their repositories"
        )
        parser.add_argument(
            "source",
            nargs="*",
            help="Source images to update in repo:tag format. "
            "Will update all source images if none provided.",
        )
        parser.add_argument(
            "--profile",
            action="append",
            help="Profiles to update, can be given multiple times. Defaults to all profiles.",
        )
        parser.add_argument(
            "--platform",
            action="append",
            help="Platforms to build, may be specified multiple times. Defaults to all platforms",
        )
        parser.add_argument(
            "--check",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Only verify that all source images are up to date",
        )
        parser.add_argument(
            "--clear",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Clear any existing cached source image digests",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        if args.clear:
            tplbld.build_data.source.clear()

        profiles = args.profile or list(tplbld.config.profiles)
        platforms = args.platform or tplbld.config.platforms

        all_source_images = set()
        for profile in profiles:
            for platform in platforms:
                stage_mapping = await tplbld.render(profile=profile, platform=platform)
                for stage_data in stage_mapping.values():

                    def _add_source_image(img: ImageDefinition) -> ImageDefinition:
                        if isinstance(img, SourceImage):
                            all_source_images.add((img.repo, img.tag, img.platform))
                        return img

                    visit_graph(
                        [stage_data.image],
                        _add_source_image,
                    )

        source_images = set()
        for image in args.source:
            image_ref = parse_image_name(image)

            found_one = False
            for platform in platforms:
                img = (image_ref.name(include_ref=False), image_ref.ref, platform)
                if img in all_source_images:
                    found_one = True
                    source_images.add(img)
            if not found_one:
                raise TplBuildException(f"No source image {image_ref} referenced")

        if not source_images:
            source_images = all_source_images

        updates_found = False
        for repo, tag, platform in source_images:
            image = SourceImage(repo=repo, tag=tag, platform=platform)
            try:
                prev_digest = (
                    await tplbld.resolve_image(image, check_only=True)
                ).digest
            except TplBuildNoSourceImageException:
                prev_digest = None

            new_digest = (
                await tplbld.resolve_image(
                    SourceImage(repo=repo, tag=tag, platform=platform),
                    force_update=True,
                )
            ).digest

            prev_digest = prev_digest or "<missing>"
            new_digest = new_digest or "<missing>"
            if new_digest == prev_digest:
                print(f"No update for {repo}:{tag} on {platform}")
            else:
                updates_found = True
                if args.check:
                    print(f"Found update for {repo}:{tag} on {platform}")
                else:
                    print(f"Updated {repo}:{tag} on {platform}")
                print(f"  {prev_digest} -> {new_digest[:32]}")

        if args.check:
            return 1 if updates_found else 0

        tplbld.save_build_data()
        return 0



from typing import Callable, Dict, Iterable, List, Optional, Tuple

from .exceptions import TplBuildException
from .hashing import json_hash
from .images import BaseImage, ImageDefinition

VisitFunc = Callable[[ImageDefinition], ImageDefinition]
VisitFuncPost = Callable[[ImageDefinition], None]


def visit_graph(
    roots: Iterable[ImageDefinition],
    visit_func: VisitFunc,
    *,
    visit_func_post: Optional[VisitFuncPost] = None,
) -> List[ImageDefinition]:
    """
    Traverses the image graph from the passed `roots` calling `visit_func` with
    each image. The graph is traveresed in pre-order allowing `visit_func` to
    change a node before it is traversed. `visit_func_post` can optionally
    also be provided. If given this will be called with each image in post-order,
    i.e. after each dependant image has been processed.

    visit_graph will raise TplBuildException if a cycle is detected in the graph.
    It will also automatically skip images taht have been visited already.
    visit_graph will return a list of the update image roots in the same order
    they were passed to visit_graph.

    `visit_func` can return a new image object (or modify and return the passed
    image object) to modify the graph. Alternatively, `visit_func` may raise a
    StopIteration exception to not traverse further into the part of the graph
    rooted at the current image.
    """

    roots_list = list(roots)
    stack: List[
        Tuple[Optional[ImageDefinition], Optional[List[ImageDefinition]], int]
    ] = [(None, roots_list, 0)]
    on_stack = set()
    remapped = {}
    while True:
        image, image_deps, dep_idx = stack[-1]
        if image_deps is None:
            assert image is not None
            try:
                new_image = visit_func(image)
            except StopIteration:
                stack.pop()
                continue

            # Store the remapping
            remapped[image] = new_image
            image = new_image
            image_deps = image.get_dependencies()
            on_stack.add(image)

            # Update dependant for parent image if not root.
            if len(stack) > 1:
                assert stack[-2][1] is not None
                stack[-2][1][stack[-2][2] - 1] = image

        dep_image = None
        while dep_image is None and dep_idx < len(image_deps):
            dep_image = image_deps[dep_idx]
            if dep_image in on_stack:
                raise TplBuildException("Cycle detected in graph")

            # Check if we've visited this image before and apply its remapping.
            remapped_image = remapped.get(dep_image)
            if remapped_image is not None:
                image_deps[dep_idx] = remapped_image
                dep_image = None

            dep_idx += 1

        stack[-1] = (image, image_deps, dep_idx)
        if dep_image is None:
            if len(stack) == 1:
                return roots_list

            assert image is not None
            image.set_dependencies(image_deps)
            on_stack.remove(image)
            stack.pop()
            if visit_func_post is not None:
                visit_func_post(image)
        else:
            stack.append((dep_image, None, 0))


def hash_graph(
    roots: Iterable[ImageDefinition],
    *,
    salt: str = "",
    symbolic=True,
) -> Dict[ImageDefinition, str]:
    """
    Generate a mapping of image nodes to their hashes.

    Arguments:
        roots: The root of the build graph to hash.
        symbolic: If set the hash is computed symbolicly. Otherwise a full
            hash is done (in particular this means build contexts will be
            fully read and hashed).
    """
    hash_mapping: Dict[ImageDefinition, str] = {}

    def hash_node(image: ImageDefinition) -> None:
        if isinstance(image, BaseImage) and image.image is not None:
            # Unwrap BaseImage references
            hash_mapping[image] = hash_mapping[image.image]
        else:
            hash_mapping[image] = json_hash(
                [
                    salt,
                    type(image).__name__,
                    image.local_hash_data(symbolic),
                    *(hash_mapping[dep] for dep in image.get_dependencies()),
                ]
            )

    visit_graph(roots, lambda image: image, visit_func_post=hash_node)
    return hash_mapping



import dataclasses
import logging
import os
from typing import Any, Dict, List, Optional

from aioregistry import parse_image_name

from .config import StageConfig, TplContextConfig
from .context import BuildContext
from .exceptions import TplBuildException, TplBuildTemplateException
from .graph import visit_graph
from .images import (
    BaseImage,
    CommandImage,
    ContextImage,
    CopyCommandImage,
    ImageDefinition,
    ScratchImage,
    SourceImage,
    StageData,
    StageDescriptor,
)
from .tplbuild import TplBuild
from .utils import (
    extract_command_flags,
    format_command_with_flags,
    ignore_escape,
    line_reader,
)

LOGGER = logging.getLogger(__name__)

RESERVED_STAGE_NAMES = {"scratch"}
SIMPLE_COMMANDS = {
    "RUN",
    "CMD",
    "LABEL",
    "MAINTAINER",
    "EXPOSE",
    "ENV",
    "ENTRYPOINT",
    "VOLUME",
    "USER",
    "WORKDIR",
    "ARG",
    "ONBUILD",
    "STOPSIGNAL",
    "HEALTHCHECK",
    "SHELL",
}


@dataclasses.dataclass(eq=False)
class _LateImageReference(ImageDefinition):
    """
    Represents a symbolic reference to another image. These should be
    resolved to other ImageDefinition objects before render returns
    any image graph data.
    """

    image_name: str

    def local_hash_data(self, symbolic: bool) -> str:
        raise NotImplementedError(
            "LateImageReference should be removed before attempting to hash"
        )


def _render_context(
    tplbld: TplBuild,
    context_config: TplContextConfig,
    profile: str,
    render_vars: Dict[str, Any],
    stage_desc: StageDescriptor,
) -> ContextImage:
    """
    Renders a context config into a ContextImage graph representation.
    """
    if stage_desc.name in RESERVED_STAGE_NAMES:
        raise TplBuildException(
            f"Cannot name context {repr(stage_desc.name)}, name is reserved"
        )

    ignore_data = context_config.ignore
    if ignore_data is None:
        ignore_file = context_config.ignore_file or ".dockerignore"
        try:
            with open(
                os.path.join(tplbld.base_dir, ignore_file), encoding="utf-8"
            ) as fign:
                ignore_data = fign.read()
        except IOError as exc:
            if context_config.ignore_file is not None:
                raise TplBuildException(f"Failed to read ignore file: {exc}") from exc
            ignore_data = ""

    try:
        ignore_data = tplbld.jinja_render(
            ignore_data,
            dict(
                platform=stage_desc.platform,
                user_config=tplbld.user_config,
                profile=profile,
                vars=render_vars,
            ),
            file_env=True,
        )
    except TplBuildTemplateException as exc:
        exc.update_message(
            f"Failed to render ignore context for {repr(stage_desc.name)}: {exc}"
        )
        raise

    ignore_lines = ignore_data.split("\n")
    ignore_lines.extend(
        ignore_escape(os.path.join(context_config.base_dir, path))
        for path in (
            *(
                os.path.join(search_path, tplbld.config.template_entrypoint)
                for search_path in tplbld.config.template_paths
            ),
            context_config.ignore_file or ".dockerignore",
            "tplbuild.yml",
            ".tplbuilddata.json",
            ".tplbuildconfig.yml",
        )
    )

    build_context = BuildContext(
        os.path.join(tplbld.base_dir, context_config.base_dir),
        None if context_config.umask is None else int(context_config.umask, 8),
        ignore_lines,
    )

    return ContextImage(
        stage_descs={stage_desc},
        context=build_context,
        platform=stage_desc.platform,
    )


def _resolve_late_references(stages: Dict[str, StageData], platform: str) -> None:
    """
    Update all the images in `stages` to remove any _LateImageReference
    objects and replace them with the proper image.
    """

    def _visit(image: ImageDefinition) -> ImageDefinition:
        """
        If visiting a late image reference replace it with the proper
        stage image or source image.
        """
        if not isinstance(image, _LateImageReference):
            return image

        stage = stages.get(image.image_name)
        if stage is not None:
            return stage.base_image or stage.image

        if image.image_name == "scratch":
            return ScratchImage(platform=platform)

        try:
            image_ref = parse_image_name(image.image_name)
        except ValueError as exc:
            raise TplBuildException(
                f"Malformed image name {repr(image.image_name)}"
            ) from exc

        repo_name = "/".join(image_ref.repo)
        if image_ref.registry:
            repo_name = f"{image_ref.registry}/{repo_name}"
        return SourceImage(
            repo=repo_name,
            tag=image_ref.ref,
            platform=platform,
        )

    stage_images = visit_graph(
        (stage_data.image for stage_data in stages.values()),
        _visit,
    )
    for stage_data, stage_image in zip(stages.values(), stage_images):
        stage_data.image = stage_image


def render(
    tplbld: TplBuild, profile: str, render_vars: Dict[str, Any], platform: str
) -> Dict[str, StageData]:
    """
    Renders all build contexts and stages into its graph representation.
    """

    def make_stage_desc(name: str) -> StageDescriptor:
        return StageDescriptor(
            name=name,
            profile=profile,
            platform=platform,
        )

    result = {
        context_name: StageData(
            name=context_name,
            image=_render_context(
                tplbld,
                context_config,
                profile,
                render_vars,
                make_stage_desc(context_name),
            ),
            config=StageConfig(),
        )
        for context_name, context_config in tplbld.config.contexts.items()
    }

    # Determine the default context as either the context named "default"
    # or the first context listed. If there are no contexts default_context
    # will just be None.
    default_context = None
    if "default" in result:
        default_context = result["default"].image
    elif result:
        default_context = next(iter(result.values())).image

    dockerfile_data, dockerfile_source_mapping = tplbld.jinja_render_debug(
        tplbld.config.template_entrypoint,
        dict(
            platform=platform,
            user_config=tplbld.user_config,
            profile=profile,
            vars=render_vars,
        ),
        file_template=True,
        file_env=True,
    )

    @dataclasses.dataclass
    class ActiveImage:
        """
        Tracks metadata on an active image in the image stack.
        """

        name: str
        image: ImageDefinition
        contexts: List[Optional[ImageDefinition]] = dataclasses.field(
            default_factory=lambda: [default_context]
        )

    image_stack: List[ActiveImage] = []

    def _pop_image_stack():
        """
        Pop the image on the top of the stack and add the stage data to the
        result.
        """
        img = image_stack.pop()
        if img.name in result:
            raise TplBuildException(f"Duplicate stage names {repr(img.name)}")

        stage_data = StageData(
            name=img.name,
            image=img.image,
            config=tplbld.get_stage_config(img.name, profile, platform, render_vars),
        )
        if stage_data.config.base:
            stage_data.base_image = BaseImage(
                profile=profile,
                stage=img.name,
                platform=platform,
                image=img.image,
            )
        result[img.name] = stage_data

    for line_pos, _, line in line_reader(dockerfile_data):
        line_num = dockerfile_source_mapping.get_source_line(line_pos)
        line_parts = line.split(maxsplit=1)
        cmd = line_parts[0].upper()
        line = line_parts[1] if len(line_parts) > 1 else ""

        if cmd == "FROM":
            line_parts = line.split()
            if len(line_parts) == 1:
                raise TplBuildException(
                    f"{line_num}: FROM without a stage name not supported"
                )
            if len(line_parts) != 3 or line_parts[1].upper() != "AS":
                raise TplBuildException(
                    f"{line_num}: Expected 'FROM parent AS stage_name'"
                )
            image_stack.append(
                ActiveImage(
                    image=_LateImageReference(line_parts[0]),
                    name=line_parts[2],
                )
            )
        elif cmd == "END":
            if line:
                raise TplBuildException(
                    f"{line_num}: Unexpected extra data after END command"
                )
            _pop_image_stack()
        elif cmd in SIMPLE_COMMANDS:
            if not image_stack:
                raise TplBuildException(f"{line_num}: Expected image start, not {cmd}")
            image_stack[-1].image = CommandImage(
                stage_descs={make_stage_desc(image_stack[-1].name)},
                parent=image_stack[-1].image,
                command=cmd,
                args=line,
            )
        elif cmd in ("ADD", "COPY"):
            if cmd == "ADD":
                LOGGER.warning(
                    "%s: Treating unsupported 'ADD' command like 'COPY'", line_num
                )
            if not image_stack:
                raise TplBuildException(f"{line_num}: Expected image start, not {cmd}")

            ctx = image_stack[-1].contexts[-1]

            line, flags = extract_command_flags(line)
            if from_name := flags.pop("from", None):
                ctx = _LateImageReference(from_name)

            line = format_command_with_flags(line, flags)
            if ctx is None:
                raise TplBuildException(f"{line_num}: Cannot COPY from null context")

            assert not isinstance(ctx, str)
            image_stack[-1].image = CopyCommandImage(
                stage_descs={make_stage_desc(image_stack[-1].name)},
                parent=image_stack[-1].image,
                context=ctx,
                command=line,
            )
        elif cmd == "PUSHCONTEXT":
            if not image_stack:
                raise TplBuildException(f"{line_num}: Expected image start, not {cmd}")
            image_stack[-1].contexts.append(_LateImageReference(line))
        elif cmd == "POPCONTEXT":
            if not image_stack:
                raise TplBuildException(f"{line_num}: Expected image start, not {cmd}")
            if len(image_stack[-1].contexts) <= 1:
                raise TplBuildException(f"{line_num}: No context on stack to pop")
            image_stack[-1].contexts.pop()
        else:
            raise TplBuildException(
                f"{line_num}: Unsupported build command {repr(cmd)}"
            )

    while image_stack:
        _pop_image_stack()

    _resolve_late_references(result, platform)

    return result



import hashlib
import json

# The default hashing algorithm used by tplbuild.
HASHER = hashlib.sha256


class HashWriter:
    """
    File-like writable object that passed all writes through to the supplied
    hasher. It will automatically encode str data using the supplied encoding.
    """

    def __init__(self, hsh, *, encoding="utf-8") -> None:
        self.hsh = hsh
        self.encoding = encoding

    def write(self, data):
        """Write data to the underlying hasher"""
        if isinstance(data, bytes):
            self.hsh.update(data)
        else:
            self.hsh.update(data.encode(self.encoding))


def json_hash(data) -> str:
    """
    Generate a crytographic hash of JSON-able data. Returns the hex digest.
    """
    hsh = HASHER()
    json.dump(data, HashWriter(hsh), sort_keys=True)  # type: ignore
    return hsh.hexdigest()



import functools
import io
import os.path
import re
import stat
import tarfile
from typing import Callable, Dict, Iterable, List, Optional, Tuple

from . import hashing
from .exceptions import TplBuildContextException, TplBuildException


@functools.lru_cache(maxsize=2**16)
def _hash_file(path: str) -> str:
    """
    Hash the passed file, cache the result.
    """
    hsh = hashing.HASHER()
    with open(path, "rb") as fdata:
        while data := fdata.read(2**16):
            hsh.update(data)
    return hsh.hexdigest()


def _create_pattern_part(
    path_pat: str, *, allow_double_star: bool = True
) -> Tuple[str, bool]:
    """
    Returns (regex_pattern, simple) tuple where `regex_pattern` is a regex that
    recognizes the passed pattern and `simple` indicates if the pattern is a
    "simple" pattern, i.e. matching only a single literal.

    `path_pat` is to be interpretted as described in
    https://pkg.go.dev/path/filepath#Match except that "**" should match any
    number of directories.

    Raises a ValueError if path_pat is malformed.
    """
    assert os.path.sep not in path_pat
    if path_pat == "**" and allow_double_star:
        # Match any number of directories
        return ".*", False

    result = [re.escape(os.path.sep)]
    simple = True

    i = 0
    while i < len(path_pat):
        ch = path_pat[i]
        i += 1

        if ch == "\\":
            if i == len(path_pat):
                raise ValueError("Trailing escape character")
            result.append(re.escape(path_pat[i]))
            i += 1
        elif ch in "*?":
            simple = False
            result.append(f"[^{os.path.sep}]{ch}")
        elif ch == "[":
            simple = False
            range_start = None
            cclass_empty = True
            char_avail = False
            result.append("[")

            # Check for character class negation
            if i < len(path_pat) and path_pat[i] == "^":
                result.append("^")
                i += 1

            while True:
                if i == len(path_pat):
                    raise ValueError("Unclosed character class")

                ch = path_pat[i]
                i += 1

                if ch == "\\":
                    if i == len(path_pat):
                        raise ValueError("Trailing escape character")
                    ch = path_pat[i]
                    i += 1
                elif ch == "]":
                    if range_start is not None:
                        raise ValueError("Unclosed character range")
                    if cclass_empty:
                        raise ValueError("Empty character class")
                    break
                elif ch == "-":
                    if not char_avail:
                        raise ValueError("Unexpected '-' in character class")
                    range_start = result[-1]
                    result.append("-")
                    char_avail = False
                    continue

                if range_start is not None:
                    if ord(range_start) > ord(ch):
                        raise ValueError("Invalid character range")
                    range_start = None
                else:
                    char_avail = True
                result.append(re.escape(ch))
                cclass_empty = False

            result.append("]")
        else:
            result.append(re.escape(ch))

    return "".join(result), simple


def _create_pattern(
    path_pat: str, match_prefix: bool, *, allow_double_star: bool = True
) -> str:
    """
    Compile a full path pattern with separators into a regex.

    If `match_prefix` is True, and all but the last component of `path_pat`
    is simple, then any path that matches a prefix of path components will also
    be matched by this pattern. For example "a/b/c/*.txt" will match "a", "a/b",
    and "a/b/c". If the pattern was instead "a/*/c/*.txt" then no prefix matching
    will happen at all because the second component is not simple.
    """
    pattern_parts = [
        _create_pattern_part(path_part, allow_double_star=allow_double_star)
        for path_part in path_pat.split(os.path.sep)
    ]
    if not match_prefix or not all(simple for _, simple in pattern_parts[:-1]):
        return (
            "^"
            + "".join(pat for pat, _ in pattern_parts)
            + f"(?:$|{re.escape(os.path.sep)})"
        )

    result = ["^"]
    for pat_part, _ in pattern_parts:
        result.append(pat_part)
        result.append("(?:$|")
    result.append(re.escape(os.path.sep))

    result.append(")" * len(pattern_parts))
    return "".join(result)


class ContextPattern:
    """
    Represents a pattern used to control what files are availble in a
    build context.

    Attributes:
        ignoring (bool): Flag indicating if matching this pattern means the
            matched element should be ignored or not ignored.
    """

    def __init__(self, pattern: str):
        try:
            if pattern.startswith("!"):
                self.ignoring = False
                self.pattern = re.compile(_create_pattern(pattern[1:], True))
            else:
                self.ignoring = True
                self.pattern = re.compile(_create_pattern(pattern, False))
        except ValueError as exc:
            raise TplBuildContextException(
                f"Error handling {repr(pattern)}: {exc}"
            ) from exc

    def matches(self, path: str) -> bool:
        """Returns True if this pattern matches the path"""
        return bool(self.pattern.search(path))


def _apply_umask(mode: int, umask: Optional[int]) -> int:
    """
    Copy the user permission bits to group and all bits and then apply
    the supplied umask. If umask is None just return mode instead.
    """
    if umask is None:
        return mode
    umode = (mode >> 6) & 0o7
    mode &= ~0o777
    mode |= ((umode << 6) | (umode << 3) | umode) & ~umask
    return mode


def _stat_to_tarinfo(
    base_path: str, arch_path: str, *, umask: Optional[int] = None, follow_link=True
) -> tarfile.TarInfo:
    """
    Convert a stat_result into a TarInfo structure.
    """
    tarinfo = tarfile.TarInfo()
    if follow_link:
        statres = os.stat(os.path.join(base_path, arch_path))
    else:
        statres = os.lstat(os.path.join(base_path, arch_path))

    linkname = ""
    stmd = statres.st_mode
    if stat.S_ISREG(stmd):
        typ = tarfile.REGTYPE
    elif stat.S_ISDIR(stmd):
        typ = tarfile.DIRTYPE
    elif stat.S_ISFIFO(stmd):
        typ = tarfile.FIFOTYPE
    elif stat.S_ISLNK(stmd):
        typ = tarfile.SYMTYPE
        linkname = os.readlink(os.path.join(base_path, arch_path))
    elif stat.S_ISCHR(stmd):
        typ = tarfile.CHRTYPE
    elif stat.S_ISBLK(stmd):
        typ = tarfile.BLKTYPE
    else:
        raise TplBuildException("Unsupported file mode in context")

    if arch_path == ".":
        tarinfo.name = "/"
    elif arch_path.startswith("./"):
        tarinfo.name = arch_path[1:]
    else:
        tarinfo.name = "/" + arch_path
    tarinfo.mode = _apply_umask(stmd, umask)
    tarinfo.uid = 0
    tarinfo.gid = 0
    tarinfo.uname = "root"
    tarinfo.gname = "root"
    if typ == tarfile.REGTYPE:
        tarinfo.size = statres.st_size
    else:
        tarinfo.size = 0
    tarinfo.mtime = 0
    tarinfo.type = typ
    tarinfo.linkname = linkname

    if typ in (tarfile.CHRTYPE, tarfile.BLKTYPE):
        if hasattr(os, "major") and hasattr(os, "minor"):
            tarinfo.devmajor = os.major(statres.st_rdev)
            tarinfo.devminor = os.minor(statres.st_rdev)
    return tarinfo


class BuildContext:
    """
    Class representing and capable of writing a build context.

    Args:
        base_dir: The base directory of the build context
        umask: If not None, the user permission bits will be copied to the
            'group' and 'all' bits and then the umask will be applied. If
            None then the exact file permissions will be forwarded to the
            build context.
        ignore_patterns: An interable of ignore patterns in the order they
            should be tested. A path will be ignored if the last pattern it
            matches in the list is not negated. This is meant to mirror the
            behavior and semantics of
            https://docs.docker.com/engine/reference/builder/#dockerignore-file
    """

    def __init__(
        self,
        base_dir: Optional[str],
        umask: Optional[int],
        ignore_patterns: Iterable[str],
    ) -> None:
        self.base_dir = base_dir
        self.umask = umask
        self.context_patterns = tuple(
            ContextPattern(pattern.strip())
            for pattern in ignore_patterns
            if pattern.strip() and pattern.strip()[0] != "#"
        )

    def ignored(self, path: str):
        """
        Returns True if the given path should be ignored (not present) in
        the build contxt. `path` should start with a directory separator
        and should be relative to `self.base_dir`.
        """
        ignored = False
        for pattern in self.context_patterns:
            if pattern.ignoring == ignored:
                continue
            if pattern.matches(path):
                ignored = pattern.ignoring
        return ignored

    def walk_context(
        self,
        *,
        extra_files: Optional[Dict[str, Tuple[int, bytes]]] = None,
        ignore_func: Optional[Callable[[str], bool]] = None,
    ) -> Iterable[tarfile.TarInfo]:
        """
        Generator that yields TarInfo objects for each not-ignored object
        in the context. Objects are yielded in a deterministic order based on the
        names.
        """

        def _is_ignored(path: str) -> bool:
            if self.ignored(path):
                return True
            return ignore_func is not None and ignore_func(path)

        if self.base_dir is None:
            tarinfo = tarfile.TarInfo("/")
            tarinfo.mode = _apply_umask(0o777, self.umask)
            tarinfo.type = tarfile.DIRTYPE
            yield tarinfo
        else:
            for root, dir_names, file_names in os.walk(self.base_dir):
                arch_root = os.path.relpath(root, self.base_dir)
                tarinfo = _stat_to_tarinfo(self.base_dir, arch_root, umask=self.umask)
                yield tarinfo

                file_names[:] = sorted(
                    file_name
                    for file_name in file_names
                    if not _is_ignored(os.path.join(tarinfo.name, file_name))
                )
                dir_names[:] = sorted(
                    dir_name
                    for dir_name in dir_names
                    if not _is_ignored(os.path.join(tarinfo.name, dir_name))
                )

                for file_name in file_names:
                    yield _stat_to_tarinfo(
                        self.base_dir,
                        os.path.join(arch_root, file_name),
                        umask=self.umask,
                        follow_link=False,
                    )

        extra_files = extra_files or {}
        for file_name, (file_mode, file_data) in extra_files.items():
            tarinfo = tarfile.TarInfo(file_name)
            tarinfo.mode = _apply_umask(file_mode, self.umask)
            tarinfo.size = len(file_data)
            yield tarinfo

    def write_context(
        self,
        io_out: io.BytesIO,
        *,
        extra_files: Optional[Dict[str, Tuple[int, bytes]]] = None,
        compress: bool = False,
    ) -> None:
        """
        Write the context to `io_out`.

        Args:
            io_out: The file-like object to write the build context to as a tar file.
            extra_files: Extra file data to add at the root of the archive.
                this is of the form (file mode, file data).
            compress: If set the output stream will be gzipped.
        """
        extra_files = extra_files or {}

        with tarfile.open(
            fileobj=io_out,
            format=tarfile.PAX_FORMAT,
            mode=("w|gz" if compress else "w|"),
        ) as tf:
            for tarinfo in self.walk_context(extra_files=extra_files):
                if tarinfo.type == tarfile.REGTYPE:
                    extra_data = extra_files.get(tarinfo.name)
                    if extra_data:
                        tf.addfile(tarinfo, fileobj=io.BytesIO(extra_data[1]))
                    else:
                        assert self.base_dir is not None
                        with open(
                            os.path.join(self.base_dir, "." + tarinfo.name), "rb"
                        ) as fileobj:
                            tf.addfile(tarinfo, fileobj=fileobj)
                else:
                    tf.addfile(tarinfo)

    def compute_partial_hash(
        self,
        *,
        patterns: Optional[List[str]] = None,
    ) -> str:
        """
        Compute a partial hash of the context where all hashed files must match
        at least one file pattern in `patterns`.
        """
        pats = [
            re.compile(_create_pattern(pat, True, allow_double_star=False))
            for pat in (patterns or [])
        ]

        def _ignore_func(path: str) -> bool:
            return not any(pat.search(path) for pat in pats)

        hsh = hashing.HASHER()
        for tarinfo in self.walk_context(
            ignore_func=_ignore_func if patterns else None
        ):
            info = tarinfo.get_info()
            info["type"] = info["type"].decode("utf-8")  # type: ignore
            hsh.update(hashing.json_hash(info).encode("utf-8"))
            if tarinfo.type == tarfile.REGTYPE:
                assert self.base_dir is not None
                hsh.update(
                    _hash_file(os.path.join(self.base_dir, "." + tarinfo.name)).encode(
                        "utf-8"
                    )
                )

        return hashing.json_hash(
            [
                type(self).__name__,
                "full",
                hsh.hexdigest(),
            ]
        )

    @functools.cached_property
    def full_hash(self) -> str:
        """The full content hash of the build context, as a hex digest"""
        return self.compute_partial_hash()

    @functools.cached_property
    def symbolic_hash(self) -> str:
        """
        The symbolic content hash of the build context, as a hex digest. This
        is different from :attr:`full_hash` in that it does not read any files
        from the build context and is only a hash of the parameters that define
        the build context instead.
        """
        return hashing.json_hash(
            [
                type(self).__name__,
                "symbolic",
                self.umask,
                self.base_dir,
                [[pat.ignoring, pat.pattern.pattern] for pat in self.context_patterns],
            ]
        )



import sys


def _extract_jinja_frames(exc_tb) -> str:
    """
    Extract all the frames in the traceback that look like jinja frames

    Returns:
        A multiline string with a formatted traceback of all the Jinja
        synthetic frames or an empty string if none were found.
    """
    lines = []
    while exc_tb:
        code = exc_tb.tb_frame.f_code
        if code.co_name in (
            "template",
            "top-level template code",
        ) or code.co_name.startswith("block "):
            lines.append(f"  at {code.co_filename}:{exc_tb.tb_lineno}")
        exc_tb = exc_tb.tb_next
    return "\n".join(lines)


class TplBuildException(Exception):
    """
    Base class of all exceptions raised by tplbuild.
    """

    def __init__(self, message: str, *, more_message: str = "") -> None:
        super().__init__(message)
        self.more_message = more_message

    def update_message(self, message: str) -> None:
        """Change the exception message"""
        self.args = (message,)


class TplBuildTemplateException(TplBuildException):
    """
    Exception due to a failure to render a template. Will automatically attach
    jinja template exception information if present in the currently active
    exception info.
    """

    def __init__(self, message) -> None:
        more_message = ""
        exc_type, exc_val, exc_tb = sys.exc_info()
        if exc_type:
            more_message = (
                f"{exc_type.__name__}: {exc_val}\n{_extract_jinja_frames(exc_tb)}"
            )
        super().__init__(message, more_message=more_message)


class TplBuildContextException(TplBuildException):
    """
    Exception during processing of build context data.
    """


class TplBuildNoSourceImageException(TplBuildException):
    """
    Exception raised when a source image lookup fails with check_only=True.
    """



import pytest

from tplbuild.utils import (
    compute_extra_vars,
    deep_merge_json,
    extract_command_flags,
    format_command_with_flags,
    ignore_escape,
    line_reader,
)


@pytest.mark.unit
def test_line_reader():
    """Test line reader behavior"""

    assert list(line_reader("hello \\")) == [(0, 0, "hello")]
    assert list(line_reader("hi\n\nthere")) == [(0, 0, "hi"), (4, 2, "there")]
    assert list(line_reader("hi\n\nthere\n")) == [(0, 0, "hi"), (4, 2, "there")]
    assert list(line_reader("hi\n\nthe\\\nre\n")) == [(0, 0, "hi"), (4, 2, "there")]
    assert list(line_reader("hi\n\nthere\\\n")) == [(0, 0, "hi"), (4, 2, "there")]
    assert list(line_reader("hi\n# comment\nthere\\\n")) == [
        (0, 0, "hi"),
        (13, 2, "there"),
    ]
    assert list(line_reader("hi\n# comment\\\nthere\\\n")) == [
        (0, 0, "hi"),
        (14, 2, "there"),
    ]
    assert list(line_reader("hi\nthere\\\n# comment")) == [
        (0, 0, "hi"),
        (3, 1, "there"),
    ]
    assert not list(line_reader(""))
    assert not list(line_reader(" #comment\n# comment\n  #  comment 2\n\n"))
    assert list(line_reader("\n\nhi\n\nthere")) == [(2, 2, "hi"), (6, 4, "there")]
    assert list(line_reader("  \\\n\nhi\n\nthere")) == [(5, 2, "hi"), (9, 4, "there")]
    assert list(line_reader("\\\nhi\n\nthere")) == [(0, 0, "hi"), (6, 3, "there")]
    assert list(line_reader("hi \\\n # comment\nthere")) == [(0, 0, "hi there")]
    assert list(line_reader("hi \\\n # comment \\\nthere")) == [(0, 0, "hi there")]
    assert not list(line_reader("\n\n\n"))


@pytest.mark.unit
@pytest.mark.parametrize(
    "line,exp_new_line,exp_flags,exp_format_line",
    [
        pytest.param(" hello ", " hello ", {}, " hello ", id="noflags"),
        pytest.param(
            "--doh=reh hello", "hello", {"doh": "reh"}, "--doh=reh hello", id="oneflag"
        ),
        pytest.param(
            "\t \v--foo=bar --bar=baz \t hello there! ",
            "hello there! ",
            {"foo": "bar", "bar": "baz"},
            "--foo=bar --bar=baz hello there! ",
            id="twoflags",
        ),
        pytest.param(
            "--bar=baz    --foo=bar   hello",
            "hello",
            {"foo": "bar", "bar": "baz"},
            "--bar=baz --foo=bar hello",
            id="twoflagsrev",
        ),
        pytest.param(
            "  --foo=bar --foo=baz hello",
            "hello",
            {"foo": "baz"},
            "--foo=baz hello",
            id="dupflag",
        ),
        pytest.param(
            " --only=flag ", "", {"only": "flag"}, "--only=flag", id="onlyflag"
        ),
    ],
)
def test_command_flags(line, exp_new_line, exp_flags, exp_format_line):
    """Test flag extraction and formatting behavior"""
    new_line, flags = extract_command_flags(line)
    assert exp_new_line == new_line
    assert exp_flags == flags
    assert format_command_with_flags(new_line, flags) == exp_format_line


@pytest.mark.unit
@pytest.mark.parametrize(
    "a,b",
    [
        pytest.param(r"hi\there", r"hi\\there"),
        pytest.param(r"./oh?", r"./oh\?"),
        pytest.param(r"./[].?\*/**", r"./\[\].\?\\\*/\*\*"),
    ],
)
def test_ignore_escape(a, b):
    """Test ignore_escape behavior"""
    assert ignore_escape(a) == b


@pytest.mark.unit
def test_deep_merge_simple():
    """Test edge behavior of deep merge"""
    x = {}
    y = {"a": "b"}
    z = {"c": "d"}
    o = [0, 1, 2]
    f = [3, 4, 5]
    assert deep_merge_json(x, y) is y
    assert deep_merge_json(y, x) is y
    assert deep_merge_json(y, z) == {"a": "b", "c": "d"}
    assert deep_merge_json(z, y) == {"a": "b", "c": "d"}
    assert deep_merge_json(y, o) is o
    assert deep_merge_json(o, y) is y
    assert deep_merge_json(o, f) is f
    assert deep_merge_json(f, o) is o


@pytest.mark.unit
@pytest.mark.parametrize(
    "lhs,rhs,result",
    [
        pytest.param(
            {"a": {"b": "c"}},
            {"a": {"d": "e"}},
            {"a": {"b": "c", "d": "e"}},
            id="merge dicts",
        ),
        pytest.param({"a": {"b": "c"}}, {"a": 123}, {"a": 123}, id="overwrite dicts"),
    ],
)
def test_deep_merge(lhs, rhs, result):
    """Test deep merge behavior"""
    assert deep_merge_json(lhs, rhs) == result


@pytest.mark.unit
@pytest.mark.parametrize(
    "set_args,expected",
    [
        pytest.param([], {}, id="no args"),
        pytest.param([(False, "a=b=c")], {"a": "b=c"}, id="more-equal"),
        pytest.param([(False, "a=123")], {"a": "123"}, id="int-like set"),
        pytest.param([(True, "a=123")], {"a": 123}, id="int set-json"),
        pytest.param(
            [(False, "a.b.c=x"), (False, "a.b.d=y")],
            {"a": {"b": {"c": "x", "d": "y"}}},
            id="set nested",
        ),
        pytest.param(
            [(True, 'a={"b":"c"}'), (True, 'a={"d":"e"}')],
            {"a": {"b": "c", "d": "e"}},
            id="merged",
        ),
        pytest.param(
            [(True, 'a={"b":"c"}'), (False, "a=x")], {"a": "x"}, id="overwrite"
        ),
    ],
)
def test_compute_extra_vars(set_args, expected):
    """Test compute_extra_vars behavior"""
    assert compute_extra_vars(set_args) == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "set_args,err_match",
    [
        pytest.param(
            [(False, "x")], "Invalid --set value, expected '='", id="no equal"
        ),
        pytest.param(
            [(True, "x")], "Invalid --set-json value, expected '='", id="no equal json"
        ),
        pytest.param(
            [(False, "=x")], "Cannot --set variable with empty key part", id="no key"
        ),
        pytest.param(
            [(True, "=1")],
            "Cannot --set-json variable with empty key",
            id="no key json",
        ),
        pytest.param([(True, "a={")], "Invalid --set-json value JSON", id="bad json"),
        pytest.param(
            [(False, "a..b=c")],
            "Cannot --set variable with empty key part",
            id="empty key part",
        ),
    ],
)
def test_compute_extra_vars_error(set_args, err_match):
    """Test compute_extra_vars error handling"""
    with pytest.raises(ValueError, match=err_match):
        compute_extra_vars(set_args)



import random
import sys
from typing import List, Optional


class OutputStream:
    """
    Class that manages writing to output for a single sub-process.
    """

    def __init__(self, title: str, color: Optional[bytes] = None) -> None:
        self.title = title
        self.prefix = b""
        if title:
            if color:
                self.prefix = color + title.encode("utf-8") + b"\x1b[0m: "
            else:
                self.prefix = title.encode("utf-8") + b": "

    async def __aenter__(self) -> "OutputStream":
        return self

    async def __aexit__(self, exc_typ, exc_val, exc_tb) -> None:
        await self.end(exc_typ is None)

    async def write(self, line: bytes, *, err: bool = False) -> None:
        """
        Write a single line of data to the output stream. Set err=True
        to write to the error stream instead of output stream.
        """
        stream = sys.stderr.buffer if err else sys.stdout.buffer
        stream.write(self.prefix)
        stream.write(line)
        if not line.endswith(b"\n"):
            stream.write(b"\n")
        stream.flush()

    async def end(self, success: bool) -> None:
        """
        End the output stream. If success is False buffered error content
        may be redisplayed.
        """


class OutputStreamer:
    """
    Class responsible for creating output streams from sub-commands and organizing
    how those outputs are displayed. For now this is just a single concreate
    implementation that writes output as it comes in directly to stdout/stderr.

    Arguments:
        use_color: If true ANSI color escape codes will be used to highlight the
                   titles of the output streams.
    """

    def __init__(self, *, use_color: bool = True) -> None:
        self.use_color = use_color
        self.remaining_colors: List[bytes] = []

    def _reset_colors(self) -> None:
        """
        Title colors are picked randomly with replacement using the 16-color
        ANSI codes. We intentionally avoid white/black variants giving us in
        total 12 colors.
        """
        self.remaining_colors = [
            *(f"\u001b[{i}m".encode("utf-8") for i in range(31, 37)),
            *(f"\u001b[{i};1m".encode("utf-8") for i in range(31, 37)),
        ]
        random.shuffle(self.remaining_colors)

    def start_stream(self, title: str) -> OutputStream:
        """
        Create a new output stream with the given title.
        """
        color = None
        if self.use_color:
            if not self.remaining_colors:
                self._reset_colors()
            color = self.remaining_colors.pop()
        return OutputStream(title, color)



import asyncio
import dataclasses
import logging
from typing import Optional

import pytest

from tplbuild.exit_context import (
    ScopedTaskExitStack,
    async_exit_context,
    get_exit_stack,
)


@dataclasses.dataclass
class IntWrapper:
    """Simple container for an int"""

    x: int = 0


async def spin(ctr: Optional[IntWrapper] = None):
    """Sleep forever and optionally increment ctr on cancellation"""
    try:
        while True:
            await asyncio.sleep(1)
    except asyncio.exceptions.CancelledError:
        if ctr is not None:
            ctr.x += 1
        raise


async def raise_it(exc):
    """Just raise an exception as a coroutine function"""
    raise exc


@pytest.mark.unit
async def test_cancel_coro(caplog):
    """
    Test that all coroutines get cancelled and their exceptions propagated
    """
    cancel_count = IntWrapper()

    async with ScopedTaskExitStack() as stack:
        stack.create_scoped_task(spin(cancel_count))
        stack.create_scoped_task(spin(cancel_count))
        await asyncio.sleep(0.1)
    assert cancel_count.x == 2

    with pytest.raises(asyncio.exceptions.CancelledError):
        async with ScopedTaskExitStack() as stack:
            stack.create_scoped_task(spin(cancel_count), propagate_cancel=True)
            stack.create_scoped_task(spin(cancel_count), propagate_cancel=True)
            await asyncio.sleep(0.1)
    assert cancel_count.x == 4

    with pytest.raises(Exception, match="exc2"):
        with caplog.at_level(logging.ERROR):
            assert not caplog.records
            async with ScopedTaskExitStack() as stack:
                with pytest.raises(Exception):
                    await stack.create_scoped_task(
                        raise_it(Exception("exc1")), propagate_exception=True
                    )
                stack.create_scoped_task(
                    raise_it(Exception("exc2")), propagate_exception=True
                )
                await asyncio.sleep(0.1)
            assert len(caplog.records) == 1


@pytest.mark.unit
async def test_async_exit_context(caplog):
    """
    Test async exit context functionality as a decorator
    """
    with pytest.raises(LookupError):
        get_exit_stack()

    cancel_count = IntWrapper()

    @async_exit_context
    async def work():
        stack = get_exit_stack()
        stack.create_scoped_task(spin(cancel_count))
        stack.create_scoped_task(spin(cancel_count))
        await asyncio.sleep(0.1)

    await work()
    assert cancel_count.x == 2
    with pytest.raises(LookupError):
        get_exit_stack()

    @async_exit_context
    async def work_cancel():
        stack = get_exit_stack()
        stack.create_scoped_task(spin(cancel_count), propagate_cancel=True)
        stack.create_scoped_task(spin(cancel_count), propagate_cancel=True)
        await asyncio.sleep(0.1)

    with pytest.raises(asyncio.exceptions.CancelledError):
        await work_cancel()
    assert cancel_count.x == 4
    with pytest.raises(LookupError):
        get_exit_stack()

    @async_exit_context
    async def work_prop():
        stack = get_exit_stack()
        with pytest.raises(Exception):
            await stack.create_scoped_task(
                raise_it(Exception("exc1")), propagate_exception=True
            )
        stack.create_scoped_task(raise_it(Exception("exc2")), propagate_exception=True)
        await asyncio.sleep(0.1)

    with caplog.at_level(logging.ERROR):
        with pytest.raises(Exception, match="exc2"):
            assert not caplog.records
            await work_prop()
        assert len(caplog.records) == 1
    with pytest.raises(LookupError):
        get_exit_stack()


@pytest.mark.unit
async def test_nested():
    """
    Test that nested contexts get different stacks. Ensure that all tasks
    complete on exit.
    """
    all_stacks = set()
    all_tasks = []

    @async_exit_context
    async def nest(depth):
        stack = get_exit_stack()
        assert stack not in all_stacks
        all_stacks.add(stack)

        if depth > 0:
            all_tasks.append(stack.create_scoped_task(nest(depth - 1)))

        for _ in range(10 if depth == 10 else 100000000):
            await asyncio.sleep(1e-3)
            assert stack is get_exit_stack()

    await asyncio.gather(*(nest(10) for _ in range(10)))
    assert all(task.done() for task in all_tasks)



import asyncio
import dataclasses
import functools
import json
import logging
import os
import shlex
from typing import Any, Dict, List, Optional, Tuple

import jinja2
import yaml
from aioregistry import (
    AsyncRegistryClient,
    ManifestListV2S2,
    ManifestV1,
    ManifestV2S2,
    RegistryBlobRef,
    RegistryException,
    parse_image_name,
)

from .arch import client_platform, normalize_platform, normalize_platform_string
from .config import BaseImageBuildData, BuildData, StageConfig, TplConfig, UserConfig
from .exceptions import (
    TplBuildException,
    TplBuildNoSourceImageException,
    TplBuildTemplateException,
)
from .graph import hash_graph, visit_graph
from .images import BaseImage, ImageDefinition, SourceImage, StageData
from .jinja_source_mapper import SourceMapper, SourceMapperExtension
from .output import OutputStreamer
from .plan import BuildOperation, BuildPlanner
from .utils import deep_merge_json, ignore_escape, open_and_swap

LOGGER = logging.getLogger(__name__)


class TplBuild:
    """
    Container class for all top level build operations.

    Arguments:
        base_dir: The base directory of the build
        config: The build configuration
    """

    @classmethod
    def from_path(
        cls,
        base_dir: str,
        *,
        user_config: UserConfig,
        registry_client: Optional[AsyncRegistryClient] = None,
    ) -> "TplBuild":
        """
        Create a TplBuild object from just the base directory. This will
        attempt to load a file named "tplbuild.yml" from within `base_dir`
        to load the configuration. If not found an empty configuration will
        be used.

        Arguments:
            base_dir: The base directory of the build
        """
        try:
            with open(
                os.path.join(base_dir, "tplbuild.yml"), encoding="utf-8"
            ) as fconfig:
                config = TplConfig(**yaml.safe_load(fconfig))
        except FileNotFoundError:
            LOGGER.warning("No tplbuild.yml found, using default configuration")
            config = TplConfig()
        except (IOError, ValueError, TypeError, yaml.YAMLError) as exc:
            raise TplBuildException(f"Failed to load configuration: {exc}") from exc
        return TplBuild(
            base_dir, config, user_config=user_config, registry_client=registry_client
        )

    def __init__(
        self,
        base_dir: str,
        config: TplConfig,
        *,
        user_config: Optional[UserConfig] = None,
        registry_client: Optional[AsyncRegistryClient] = None,
    ) -> None:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from .executor import BuildExecutor

        self.base_dir = base_dir
        self.config = config
        self.user_config = user_config or UserConfig()
        self.planner = BuildPlanner()
        self.custom_client = bool(registry_client)
        self.registry_client = registry_client or AsyncRegistryClient()
        self.executor = BuildExecutor(self)
        self.jinja_env = jinja2.Environment(
            undefined=jinja2.StrictUndefined,
            extensions=[SourceMapperExtension],
        )
        self.jinja_file_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                [
                    os.path.normpath(os.path.join(base_dir, path))
                    for path in config.template_paths
                ]
            ),
            extensions=[SourceMapperExtension],
            undefined=jinja2.StrictUndefined,
        )
        self.jinja_env.filters["shell_escape"] = shlex.quote
        self.jinja_env.filters["ignore_escape"] = ignore_escape
        self.jinja_file_env.filters["shell_escape"] = shlex.quote
        self.jinja_file_env.filters["ignore_escape"] = ignore_escape
        self.output_streamer = OutputStreamer()
        try:
            with open(
                os.path.join(base_dir, ".tplbuilddata.json"), encoding="utf-8"
            ) as fbuilddata:
                self.build_data = BuildData(**json.load(fbuilddata))
        except FileNotFoundError:
            LOGGER.warning(".tplbuilddata.json not found, using empty bulid data")
            self.build_data = BuildData()
        except (IOError, ValueError, TypeError, yaml.YAMLError) as exc:
            raise TplBuildException(
                f"Failed to load .tplbuilddata.json: {exc}"
            ) from exc

    async def __aenter__(self) -> "TplBuild":
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        if not self.custom_client:
            await self.registry_client.__aexit__(exc_type, exc_value, exc_traceback)

    def jinja_render_debug(
        self,
        template: str,
        params: Dict[str, Any],
        *,
        file_template=False,
        file_env=False,
    ) -> Tuple[str, SourceMapper]:
        """
        Like jinja_render but also returns a SourceMapper object that can be used to
        map output text to the template line that generated it.
        """
        jinja_env = self.jinja_file_env if file_env else self.jinja_env
        try:
            if file_template:
                jtpl = jinja_env.get_template(template)
            else:
                jtpl = jinja_env.from_string(template)

            source_mapper = jinja_env.extensions[
                "tplbuild.jinja_source_mapper.SourceMapperExtension"
            ]
            return source_mapper.render(jtpl.generate(params))  # type: ignore
        except jinja2.TemplateError as exc:
            if file_template:
                raise TplBuildTemplateException(
                    f"Failed to render file template {repr(template)}"
                ) from exc
            raise TplBuildTemplateException("Failed to render template") from exc

    def jinja_render(
        self,
        template: str,
        params: Dict[str, Any],
        *,
        file_template=False,
        file_env=False,
    ) -> str:
        """
        Renders a Jinja template in the appropriate jinja environment and returns
        the rendered result. If `file_env` is True this will allow other templates
        to be loaded using include commands. If `file_template` is True this will
        interpret `template` as a file path to a template rather than a template.

        On Failure this will raise a TplBuildTempalteException.
        """
        return self.jinja_render_debug(
            template, params, file_template=file_template, file_env=file_env
        )[0]

    def get_base_image_name(
        self,
        base_image: BaseImage,
        *,
        use_digest: bool = False,
    ) -> str:
        """
        Return the base image name by rendering the base image name
        config template.
        """
        if self.config.base_image_repo is None:
            raise TplBuildException("Must configure base_image_repo to use base images")
        try:
            base_image_repo = self.jinja_render(
                self.config.base_image_repo,
                dict(
                    stage_name=base_image.stage,
                    profile=base_image.profile,
                    platform=base_image.platform,
                ),
            )
        except TplBuildTemplateException as exc:
            exc.update_message("Failed to render base image name template")
            raise exc

        if use_digest:
            return f"{base_image_repo}@{base_image.digest}"
        return f"{base_image_repo}:{base_image.content_hash}"

    def get_stage_config(
        self, stage_name: str, profile: str, platform: str, render_vars: Dict[str, Any]
    ) -> StageConfig:
        """
        Return the StageConfig for the given stage name. This will first check
        :attr:`TplConfig.stages` to see if there is an explicit configuration
        for this stage.

        Otherwise, if the stage name starts with "base-" or "base\\_" a base
        image `StageConfig` will be returned.

        If the name starts with 'anon-' or 'anon\\_' instead the stage will be
        considered an "anonymous" stage. This means that it will neither be built
        as a base image nor will it be tagged or pushed to any image name. Such
        images are used for intermediate build operations and are not an output
        of the build themselves.

        For every other image the image_names and push_names default values will
        be substituted where needed and the templates will be evaluated.
        """
        image_names = [self.config.stage_image_name]
        push_names = [self.config.stage_push_name]
        if stage_config := self.config.stages.get(stage_name):
            if stage_config.image_names is not None:
                image_names = stage_config.image_names
            if stage_config.push_names is not None:
                push_names = stage_config.push_names
        if stage_name[0:5] in ("base-", "base_"):
            return StageConfig(base=True)
        if stage_name[0:5] in ("anon-", "anon_"):
            return StageConfig()
        params = dict(
            stage_name=stage_name,
            profile=profile,
            platform=platform,
            vars=render_vars,
        )
        try:
            image_names = [
                self.jinja_render(image_name, params) for image_name in image_names
            ]
        except TplBuildTemplateException as exc:
            exc.update_message("Failed to render stage image name template")
            raise exc

        try:
            push_names = [
                val
                for push_name in push_names
                if (val := self.jinja_render(push_name, params))
            ]
        except TplBuildTemplateException as exc:
            exc.update_message("Failed to render stage push name template")
            raise exc

        return StageConfig(
            image_names=image_names,
            push_names=push_names,
        )

    def lookup_base_image(
        self,
        stage_name: str,
        platform: str,
        *,
        profile: str = "",
    ) -> Tuple[str, BaseImage]:
        """
        Returns the full image name and unrendered :class:`BaseImage` image node
        that represents the requested base image.

        Raises:
            KeyError: If the base image does not exist or has not been built
            TplBuildException: If no base image repo is configured
        """
        if self.config.base_image_repo is None:
            raise TplBuildException("No base image repo format configured")

        profile = profile or self.default_profile
        cached_build_data = self.build_data.base[profile][stage_name][platform]
        image = BaseImage(
            profile=profile,
            stage=stage_name,
            platform=platform,
            content_hash=cached_build_data.build_hash,
            digest=cached_build_data.image_digest,
        )
        return (
            self.get_base_image_name(image, use_digest=True),
            image,
        )

    @property
    def default_profile(self) -> str:
        """
        Returns the default configuration name. If there are no configuration
        profiles then this will return an empty string.
        """
        return self.config.default_profile or next(iter(self.config.profiles), "")

    async def get_default_platform(self) -> str:
        """
        Calculate the default platform.
        """
        options = [
            await self.executor.platform(),
            client_platform(),
        ]
        for platform in options:
            if platform:
                platform = normalize_platform_string(platform)
                if platform in self.config.platforms:
                    return platform
        return self.config.platforms[0]

    def get_render_vars(
        self,
        profile: str,
        *,
        extra_vars: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Returns the render context to pass into Jinja to render the build file
        for the given profile.
        """
        if not profile:
            profile_data = {}
        else:
            try:
                profile_data = self.config.profiles[profile]
            except KeyError as exc:
                raise TplBuildException(
                    f"Profile {repr(profile)} does not exist"
                ) from exc

        return deep_merge_json(profile_data, extra_vars or {})

    async def render(
        self,
        *,
        profile: str = "",
        platform: str = "",
        extra_vars: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, StageData]:
        """
        Render all contexts and stages into StageData.

        Returns:
            dict mapping stage names/context names to their
            respective :class:`StageData`.
        """
        # pylint: disable=import-outside-toplevel,cyclic-import
        from .render import render

        profile = profile or self.default_profile
        platform = platform or await self.get_default_platform()
        render_vars = self.get_render_vars(profile, extra_vars=extra_vars)
        return render(self, profile, render_vars, platform)

    def plan(
        self,
        stages: List[StageData],
    ) -> List[BuildOperation]:
        """
        Plan how to build the given list of stages.

        Arguments:
            stages: The list of stages to build in the form of
                :class:`StageData` objects. Use :meth:`render`
                to render all the stages for a given profile.
                You should also use :meth:`resolve_source_images`
                and :meth:`resolve_base_images` before calling `plan`.

        Returns:
            A list of :class:`BuildOperation` objects that that define how
            the stages are to be built. This list is returned topologically
            sorted so that all dependencies of a stage appear before it.
            Typically this should be fed into :meth:`build`.
        """
        return self.planner.plan(stages)

    async def build(
        self,
        build_ops: List[BuildOperation],
    ) -> None:
        """
        Execute the build operations.

        Arguments:
            build_ops: The list of build operations to execute. This should be
                topologically sorted each build operation appears only after all
                of its dependencies have been listed. Typically this should just
                be the return value from :meth:`plan`.
        """

        async def complete_callback(build_op: BuildOperation, primary_tag: str) -> None:
            """
            Update our internal build data cache for base images.
            """
            # pylint: disable=unused-argument
            for stage in build_op.stages:
                if stage.base_image:
                    assert stage.base_image.content_hash is not None
                    try:
                        descriptor = await self.registry_client.ref_lookup(
                            parse_image_name(self.get_base_image_name(stage.base_image))
                        )
                    except RegistryException as exc:
                        raise TplBuildException(
                            "Could not find pushed base image in registry"
                        ) from exc
                    if descriptor is None:
                        raise TplBuildException(
                            "Could not find pushed base image in registry"
                        )

                    self.build_data.base.setdefault(
                        stage.base_image.profile, {}
                    ).setdefault(stage.base_image.stage, {})[
                        stage.base_image.platform
                    ] = BaseImageBuildData(
                        build_hash=stage.base_image.content_hash,
                        image_digest=descriptor.digest,
                    )

                # Save base images as soon as their done in case the build
                # is later interrupted.
                self.save_build_data()

        await self.executor.build(build_ops, complete_callback=complete_callback)

    async def resolve_image(
        self,
        image: SourceImage,
        *,
        check_only=False,
        force_update=False,
    ) -> SourceImage:
        """
        Resolves a SourceImage returning a new SourceImage object with the
        digest field set. If digest is already set on `image` then this will
        just return `image` itself.

        Arguments:
            image: Source image definition to resolve
            check_only: Do not attempt to fetch the image if it is not cached.
                This will trigger a TplBuildException if the image is not available.
            force_update: Update the image digest even if it is locally cached

        Raises:
            TplBuildException: if the image could not be fetched or if it
                has no matching platform.
        """
        if image.digest is not None:
            return image

        source_digest = (
            self.build_data.source.get(image.repo, {})
            .get(image.tag, {})
            .get(image.platform)
        )
        if source_digest is not None and not force_update:
            return dataclasses.replace(image, digest=source_digest)

        if check_only:
            raise TplBuildNoSourceImageException(
                "Source image does not exist, check only"
            )

        # Grab the latest digest for the given source image.
        LOGGER.info("Looking up source image %s:%s", image.repo, image.tag)
        manifest_ref = parse_image_name(f"{image.repo}:{image.tag}")
        try:
            descriptor = await self.registry_client.ref_lookup(manifest_ref)
            if descriptor is None:
                raise TplBuildException(
                    f"Failed to lookup source image {repr(str(manifest_ref))}"
                )
            manifest_ref = manifest_ref.copy(update=dict(ref=descriptor.digest))
        except RegistryException as exc:
            raise TplBuildException(
                f"Failed to lookup source image {repr(str(manifest_ref))}"
            ) from exc

        # Download the manifest
        try:
            manifest = await self.registry_client.manifest_download(manifest_ref)
        except RegistryException as exc:
            raise TplBuildException("Failed to download source manifest") from exc

        # TODO: Should probably add some support in aioregistry itself for inspecting
        #       image configs and getting platform information.
        if isinstance(manifest, ManifestV1):
            # Only one platform available, make sure it's the one we want.
            image_platform = "linux/" + manifest.architecture
            if image.platform != image_platform:
                raise TplBuildException(
                    f"Source image has incorrect architecture {repr(image_platform)}"
                )
        elif isinstance(manifest, ManifestV2S2):
            # Only one platform available, make sure it's the one we want. V2 images
            # require us to download the config to inspect.
            config_ref = RegistryBlobRef(
                registry=manifest_ref.registry,
                repo=manifest_ref.repo,
                ref=manifest.config.digest,
            )
            _, chunk_data = await self.registry_client.ref_content_stream(config_ref)
            config_raw_data = [chunk async for chunk in chunk_data]
            config = json.loads(b"".join(config_raw_data).decode("utf-8"))
            image_platform = normalize_platform(
                config["os"],
                config["architecture"],
                config.get("variant", ""),
            )
            if image.platform != image_platform:
                raise TplBuildException(
                    f"Source image has incorrect architecture {repr(image_platform)}"
                )
        elif isinstance(manifest, ManifestListV2S2):
            # Select appropriate sub-manifest based on the platform
            for sub_manifest in manifest.manifests:
                platform = sub_manifest.platform
                if platform is None:
                    continue
                image_platform = normalize_platform(
                    platform.os, platform.architecture, platform.variant
                )
                if image.platform == image_platform:
                    manifest_ref = manifest_ref.copy(
                        update=dict(ref=sub_manifest.digest)
                    )
                    break
            else:
                raise TplBuildException(
                    "No appropriate platform found in manifest list"
                )

        # Save the result into our build data
        self.build_data.source.setdefault(image.repo, {}).setdefault(image.tag, {})[
            image.platform
        ] = manifest_ref.ref
        return dataclasses.replace(image, digest=manifest_ref.ref)

    async def resolve_base_images(
        self,
        stages: List[StageData],
        *,
        dereference=False,
        resolve_from_registry=True,
    ) -> None:
        """
        Resolve all :class:`BaseImage` nodes found in the image graph, setting
        their `content_hash` attribute based on cached build data.

        Arguments:
            dereference: If True then all BaseImage nodes will be fully hashed
                (including their build contexts) to produce a content hash
                that will be set in the BaseImage node. This content hash will
                be available at `stage.base_image` for any passed stages that
                represent base images.

                Further, when base images are resolved, if the computed
                `content_hash` does not match the cached `content_hash` then
                the node will be replaced with the base image's underlying
                build definition.
            resolve_from_registry: Modifies behavior when dereferencing to first
                check if the image is present in the registry before dereferencing
                the image. If the image is present the build data will be updated
                automatically.

        This modifies the passed in build graph and does not return anything
        directly.
        """

        full_hash_mapping = {}
        if dereference:
            full_hash_mapping = await asyncio.get_running_loop().run_in_executor(
                None,
                functools.partial(
                    hash_graph,
                    (stage.image for stage in stages if stage.base_image),
                    salt=self.build_data.hash_salt,
                    symbolic=False,
                ),
            )

            for stage in stages:
                if stage.base_image:
                    stage.base_image.content_hash = full_hash_mapping[stage.image]
                    stage.image = stage.base_image

            if resolve_from_registry:
                base_images: List[BaseImage] = [
                    stage.base_image
                    for stage in stages
                    if stage.base_image and stage.base_image.image
                ]

                for base_image in base_images:
                    cached_build_data = (
                        self.build_data.base.get(base_image.profile, {})
                        .get(base_image.stage, {})
                        .get(base_image.platform)
                    )
                    cached_content_hash = (
                        cached_build_data.build_hash if cached_build_data else None
                    )

                    assert base_image.image
                    if full_hash_mapping[base_image.image] != cached_content_hash:
                        try:
                            descriptor = await self.registry_client.ref_lookup(
                                parse_image_name(self.get_base_image_name(base_image))
                            )
                        except RegistryException as exc:
                            raise TplBuildException(
                                "Failed to lookup base image in registry"
                            ) from exc

                        if descriptor:
                            self.build_data.base.setdefault(
                                base_image.profile, {}
                            ).setdefault(base_image.stage, {})[
                                base_image.platform
                            ] = BaseImageBuildData(
                                build_hash=full_hash_mapping[base_image.image],
                                image_digest=descriptor.digest,
                            )

                            LOGGER.info(
                                "Updating base image %s:%s:%s from registry",
                                base_image.stage,
                                base_image.profile,
                                base_image.platform,
                            )
                            self.save_build_data()

        def _resolve_base(image: ImageDefinition) -> ImageDefinition:
            if not isinstance(image, BaseImage):
                return image

            cached_build_data = (
                self.build_data.base.get(image.profile, {})
                .get(image.stage, {})
                .get(image.platform)
            )

            if dereference:
                if image.image is None:
                    raise TplBuildException(
                        "Attempt to dereference base image that has no image definition"
                    )

                if (
                    cached_build_data is None
                    or full_hash_mapping[image.image] != cached_build_data.build_hash
                ):
                    return image.image

            if cached_build_data is None:
                raise TplBuildException(
                    f"No cached build of {image.profile}/{image.stage}"
                )

            image.image = None
            image.content_hash = cached_build_data.build_hash
            image.digest = cached_build_data.image_digest

            return image

        images = visit_graph((stage.image for stage in stages), _resolve_base)
        for stage, image in zip(stages, images):
            stage.image = image

            # If we're really building a base image add its push name.
            if stage.base_image and stage.image is not stage.base_image:
                stage.config.push_names = stage.config.push_names or []
                stage.config.push_names.append(
                    self.get_base_image_name(stage.base_image)
                )

    async def resolve_source_images(
        self,
        stages: List[StageData],
        *,
        check_only=False,
        force_update=False,
    ) -> None:
        """
        Resolve the manifest digest for each source image in the build graph.
        This updates the build graph passed in to resolve_source_images and does
        not return anything directly.
        """
        # Find all the source images first
        source_images = set()

        def _visit_image(image: ImageDefinition) -> ImageDefinition:
            if isinstance(image, SourceImage):
                source_images.add(image)
            return image

        visit_graph((stage.image for stage in stages), _visit_image)

        # Resolve all images into self.build_data.source.
        source_image_map = dict(
            zip(
                source_images,
                await asyncio.gather(
                    *(
                        self.resolve_image(
                            image, check_only=check_only, force_update=force_update
                        )
                        for image in source_images
                    )
                ),
            )
        )
        self.save_build_data()

        # Swap out SourceImage for the resolved ExternalImages
        def _replace_source_image(image: ImageDefinition) -> ImageDefinition:
            if isinstance(image, SourceImage):
                return source_image_map[image]
            return image

        images = visit_graph((stage.image for stage in stages), _replace_source_image)
        for stage, image in zip(stages, images):
            stage.image = image
            if stage.base_image is not None:
                stage.base_image.image = image

    def save_build_data(self):
        """
        Save build data to disk.
        """
        try:
            with open_and_swap(
                os.path.join(self.base_dir, ".tplbuilddata.json"),
                mode="w",
                encoding="utf-8",
            ) as fdata:
                json.dump(
                    self.build_data.dict(),
                    fdata,
                    indent=2,
                )
        except IOError as exc:
            raise TplBuildException(f"Failed to save build data: {exc}") from exc



from typing import Iterable, List
from unittest.mock import Mock, call

import pytest

from tplbuild.exceptions import TplBuildException
from tplbuild.graph import visit_graph
from tplbuild.images import CommandImage, CopyCommandImage, ImageDefinition, SourceImage


class ImageTestNode(ImageDefinition):
    """Test node with variable number of dependencies"""

    def __init__(self, depth: int, deps: List[ImageDefinition]) -> None:
        super().__init__()
        self.depth = depth
        self.deps = deps

    def local_hash_data(self, symbolic: bool) -> str:
        raise NotImplementedError()

    def get_dependencies(self) -> List[ImageDefinition]:
        return list(self.deps)

    def set_dependencies(self, deps: Iterable[ImageDefinition]) -> None:
        self.deps = list(deps)


@pytest.mark.unit
def test_visit_graph_cycle():
    """Test that cycle detection works correct for visit_graph"""
    img = CommandImage(
        stage_descs=set(),
        parent=None,
        command="RUN",
        args="stuff",
    )
    img.parent = img

    visit_mock = Mock(side_effect=lambda img: img)
    with pytest.raises(TplBuildException, match="Cycle detected in graph"):
        visit_graph([img], visit_mock)
    visit_mock.assert_called_once_with(img)

    img_list = [SourceImage(platform="linux/amd64", repo="hi", tag="bye")]
    for _ in range(9):
        img_list.append(
            CopyCommandImage(
                stage_descs=set(),
                parent=img_list[-1],
                context=img_list[-1],
                command="hello",
            )
        )

    # Ensure that we don't detect a cycle for overlapping DAG and that
    # we only visit each unique image once.
    visit_mock.reset_mock()
    assert visit_graph(img_list, visit_mock) == img_list
    assert visit_mock.call_count == 10

    # Make sure we detect a deep cycle
    img_list[2].context = img_list[6]
    visit_mock.reset_mock()
    with pytest.raises(TplBuildException, match="Cycle detected in graph"):
        visit_graph(img_list, visit_mock)


@pytest.mark.unit
def test_visit_graph_generate():
    """Test visit graph behavior toward genearting new nodes"""

    def visit_replace(image: ImageDefinition) -> ImageDefinition:
        assert isinstance(image, ImageTestNode)

        if image.depth < 10:
            return ImageTestNode(
                image.depth,
                [ImageTestNode(image.depth + 1, []) for _ in range(2)],
            )

        return image

    root = ImageTestNode(0, [])
    result = visit_graph([root], visit_replace)
    assert len(result) == 1
    assert len(result) == 1 and result[0] is not root

    visit_mock = Mock(side_effect=lambda img: img)
    visit_graph(result, visit_mock)
    assert visit_mock.call_count == 2**11 - 1

    def visit_update(image: ImageDefinition) -> ImageDefinition:
        assert isinstance(image, ImageTestNode)

        if image.depth < 10:
            image.set_dependencies(
                [ImageTestNode(image.depth + 1, []) for _ in range(2)]
            )

        return image

    root = ImageTestNode(0, [])
    result = visit_graph([root], visit_update)
    assert len(result) == 1 and result[0] is root

    visit_mock.reset_mock()
    visit_graph(result, visit_mock)
    assert visit_mock.call_count == 2**11 - 1


@pytest.mark.unit
def test_visit_graph():
    """Test visit graph behavior"""
    nodes = []
    for _ in range(10):
        nodes.append(ImageTestNode(-1, list(nodes)))

    visit_mock = Mock(side_effect=lambda img: img)
    assert visit_graph(nodes, visit_mock) == nodes
    assert visit_mock.call_count == 10

    new_node = ImageTestNode(-1, [])

    def visit_replace(image: ImageDefinition) -> ImageDefinition:
        if image is nodes[3]:
            return new_node
        return image

    new_nodes = visit_graph(nodes, visit_replace)
    new_nodes_exp = list(nodes)
    new_nodes_exp[3] = new_node
    assert new_nodes == new_nodes_exp

    for image in nodes[4:]:
        assert nodes[3] not in image.deps
        assert new_node in image.deps


@pytest.mark.unit
def test_visit_graph_order():
    """Test visit graph order traversal"""
    nodes = [ImageTestNode(-1, []) for _ in range(5)]
    nodes[2].deps = [nodes[3]]
    nodes[1].deps = [nodes[4], nodes[2]]
    nodes[0].deps = [nodes[4], nodes[1]]

    visit_mock_pre = Mock(side_effect=lambda img: img)
    visit_mock_post = Mock()
    assert (
        visit_graph(nodes[:1], visit_mock_pre, visit_func_post=visit_mock_post)
        == nodes[:1]
    )

    visit_mock_pre.assert_has_calls(
        [
            call(nodes[0]),
            call(nodes[4]),
            call(nodes[1]),
            call(nodes[2]),
            call(nodes[3]),
        ]
    )
    visit_mock_post.assert_has_calls(
        [
            call(nodes[4]),
            call(nodes[3]),
            call(nodes[2]),
            call(nodes[1]),
            call(nodes[0]),
        ]
    )



import argparse
import sys

from tplbuild.cmd.utility import CliUtility
from tplbuild.tplbuild import TplBuild


class BaseLookupUtility(CliUtility):
    """CLI utility entrypoint for building base images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "print the full image name of the requested base images"
        parser.add_argument(
            "image",
            nargs="*",
            help="The base image stage name to lookup",
        )
        parser.add_argument(
            "--profile",
            required=False,
            help="Profile to lookup the base image for",
        )
        parser.add_argument(
            "--platform",
            required=False,
            help="Platform to lookup the base image of. Defaults to current executor platform",
        )
        parser.add_argument(
            "--tag-only",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Only print the image tag",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        """Print out all the base image names/tags requested"""
        platform = args.platform or await tplbld.get_default_platform()
        for stage_name in args.image:
            try:
                image_name, image = tplbld.lookup_base_image(
                    stage_name, platform, profile=args.profile
                )
            except KeyError:
                sys.stderr.write(f"could not find base image {repr(stage_name)}\n")
                return 1

            if args.tag_only:
                print(image.content_hash)
            else:
                print(image_name)

        return 0



import asyncio
import threading


class SyncToAsyncPipe:
    """
    Provides a pipe interface that allows synchronous writes and asyncrhonous
    reads of byte data.

    Arguments:
        loop: The event loop from which the pipe will be read. If not provided
            the current event loop will be used. It is an error to read from
            any other event loop.
        buf_max: The maximum number of bytes that that can be buffered in the pipe.
    """

    def __init__(self, loop=None, *, buf_max=2**16) -> None:
        if loop is None:
            loop = asyncio.get_running_loop()
        self.loop = loop
        self.buf = bytearray(0 for _ in range(buf_max))
        self.buf_pos = 0
        self.buf_size = 0
        self.buf_lock = threading.Condition()
        self.waiter = None
        self.closed = False

    def write(self, data: bytes) -> None:
        """
        Write `data` to the stream. This will block if the stream buffers have
        already filled up.
        """
        data_pos = 0
        with self.buf_lock:
            while data_pos < len(data):
                while self.buf_size == len(self.buf):
                    self.buf_lock.wait()

                write_offset = (self.buf_pos + self.buf_size) % len(self.buf)
                amt = min(
                    len(data) - data_pos,
                    len(self.buf) - self.buf_size,
                    len(self.buf) - write_offset,
                )

                self.buf[write_offset : write_offset + amt] = data[
                    data_pos : data_pos + amt
                ]
                self.buf_size += amt
                data_pos += amt
                self._notify_reader()

    async def read(self) -> bytes:
        """
        Read data from the stream. Only one task can read from a pipe concurrently.

        If the stream is closed returns b"". Otherwise returns a non-empty bytes
        data object.
        """
        waiter = None
        while True:
            if waiter is not None:
                # Wait for someone to write some data.
                await waiter

            with self.buf_lock:
                if self.buf_size == 0 and not self.closed:
                    # Create new waiter, release lock, and wait for notification.
                    if self.waiter is not None:
                        raise RuntimeError(
                            "Cannot read from the same pipe multiple times concurrently"
                        )

                    waiter = self.loop.create_future()
                    self.waiter = waiter
                    continue

                if self.closed and self.buf_size == 0:
                    return b""

                amt = min(self.buf_size, len(self.buf) - self.buf_pos)
                result = bytes(self.buf[self.buf_pos : self.buf_pos + amt])
                self.buf_pos = (self.buf_pos + amt) % len(self.buf)
                self.buf_size -= amt
                self.buf_lock.notify()
                return result

    def close(self) -> None:
        """
        Close the stream. This can be called from the synchronous or asynchronous
        context.

        If a thread attempts to/is currently writing to the stream it will raise
        a BrokenPipeError.

        Readers will continue to be able to read the rest of the data in the buffer
        and then get an empty b"" object when the stream is exhausted.
        """
        with self.buf_lock:
            self.closed = True
            self.buf_lock.notify()
            self._notify_reader()

    def _notify_reader(self) -> None:
        """
        Wake up any readers that are waiting for data to be written.
        """
        if self.waiter is not None:
            self.loop.call_soon_threadsafe(self.waiter.set_result, None)
            self.waiter = None



import argparse
import uuid
from typing import List

from tplbuild.cmd.common import debug_build_operations
from tplbuild.cmd.utility import CliUtility
from tplbuild.exceptions import TplBuildException
from tplbuild.images import StageData
from tplbuild.tplbuild import TplBuild


class BaseBuildUtility(CliUtility):
    """CLI utility entrypoint for building base images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "rebuild any out-of-date base images"
        parser.add_argument(
            "image",
            nargs="*",
            help="Base image stage names to build. Defaults to all base stages",
        )
        parser.add_argument(
            "--profile",
            action="append",
            help="Profile to build, can be given multiple times. Defaults to all profiles.",
        )
        parser.add_argument(
            "--platform",
            action="append",
            help="Platforms to build, may be specified multiple times. Defaults to all platforms",
        )
        parser.add_argument(
            "--update-sources",
            default=False,
            const=True,
            action="store_const",
            help="If set the digest for each source image will be updated",
        )
        parser.add_argument(
            "--update-salt",
            default=False,
            const=True,
            action="store_const",
            help="Update salt forcing base images to be rebuilt",
        )
        parser.add_argument(
            "--check",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Only verify that all requested base images are already built",
        )
        parser.add_argument(
            "--debug",
            default=False,
            const=True,
            action="store_const",
            help="Only print rendered Dockerfiles instead of building images",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        images = set(args.image)
        profiles = args.profile or list(tplbld.config.profiles)
        platforms = args.platform or tplbld.config.platforms

        if args.update_salt:
            tplbld.build_data.hash_salt = str(uuid.uuid4())

        # Render all build stages
        stages_to_build: List[StageData] = []
        for profile in profiles:
            for platform in platforms:
                stage_mapping = await tplbld.render(profile=profile, platform=platform)
                stages_to_build.extend(
                    stage_data
                    for stage_name, stage_data in stage_mapping.items()
                    if stage_data.base_image is not None
                    and (not images or stage_name in images)
                )

        if args.check and args.update_sources:
            raise TplBuildException("Cannot pass --check and --update-sources")

        # Resolve the locked source image manifest content address from cached
        # build data.
        await tplbld.resolve_source_images(
            stages_to_build,
            check_only=args.check,
            force_update=args.update_sources,
        )

        # Replace BaseImage nodes in the build graph with their underlying
        # build definition.
        await tplbld.resolve_base_images(stages_to_build, dereference=True)

        # Create a plan of build operations to execute the requested build.
        build_ops = tplbld.plan(stages_to_build)
        if not build_ops:
            print("Base images up to date!")
            return 0
        if args.check:
            print(f"Needed {len(build_ops)} build operations")
            return 1

        # Execute the build operations.
        if args.debug:
            debug_build_operations(tplbld, build_ops)
        else:
            await tplbld.build(build_ops)

        return 0



import abc
import argparse

from tplbuild.tplbuild import TplBuild


class CliUtility(metaclass=abc.ABCMeta):
    """Abstract utility base class"""

    @abc.abstractmethod
    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        """Allow the utility to set up its arg parser.

        Arguments:
            parser (argparse.ArgumentParser): the parser to setup arguments for.
        """

    @abc.abstractmethod
    async def main(self, args, tplbld: TplBuild) -> int:
        """Run utility with parsed args and the `tplbld` configuration.

        Returns:
            The utility exit code
        """



import contextlib
import io
import os
import re
import tarfile
import tempfile
from typing import Optional

import pytest

from tplbuild.context import BuildContext, _create_pattern, _create_pattern_part


@contextlib.contextmanager
def set_umask(msk):
    """Set the umask in a context manager"""
    prev_msk = os.umask(msk)
    try:
        yield
    finally:
        os.umask(prev_msk)


def make_test(base_dir: str, mode: int, testdata, path: Optional[str] = None) -> None:
    """
    Create a file hierarchy for testing.

    Arguments:
        mode: the access mode that the root file should be created with.
        testdata:
            - If a dictionary then the root file will be a directory
              with children file names given by the keys of the dictionary
              and the values being tuples of access modes and recursive
              testdata structures that indicate how to create those sub files.
            - Otherwise should be a 'bytes' object giving the data to write
              to the file.
        path: The path to write the file hiearchy. If `path` is None
              it will use the tempdir path.
    """
    if isinstance(testdata, dict):
        if path is None:
            os.chmod(base_dir, mode)
        else:
            os.mkdir(path, mode)

        path = path or base_dir
        for subfile, subfile_data in testdata.items():
            make_test(base_dir, *subfile_data, path=os.path.join(path, subfile))  # type: ignore
    else:
        assert path is not None
        assert isinstance(testdata, bytes)
        with open(os.open(path, os.O_CREAT | os.O_WRONLY, mode), "wb") as fout:
            fout.write(testdata)


@pytest.mark.unit
def test_create_pattern_part():
    """Test _create_pattern_part functionality"""
    no_sep = f"[^{re.escape(os.path.sep)}]"
    pass_tests = {
        "**": (".*", False),
        "?": (f"/{no_sep}?", False),
        "*": (f"/{no_sep}*", False),
        "hello world": (r"/hello\ world", True),
        r"hello\[world": (r"/hello\[world", True),
        "[^^]": (r"/[^\^]", False),
        "[^a-b]": ("/[^a-b]", False),
        r"[\^\]]": (r"/[\^\]]", False),
        "_[ab-yz]?.*": (f"/_[ab-yz]{no_sep}?\\.{no_sep}*", False),
        ".{ }^|": (r"/\.\{\ \}\^\|", True),
        "a[[]b]c": (r"/a[\[]b\]c", False),
    }
    value_error_tests = {
        "[hi": "Unclosed character class",
        "[hi\\]": "Unclosed character class",
        "hi\\": "Trailing escape character",
        "[hi\\": "Trailing escape character",
        "[^-c]": "Unexpected '-' in character class",
        "[a-b-c]": "Unexpected '-' in character class",
        "[b-a]": "Invalid character range",
        "[]": "Empty character class",
        "[^]": "Empty character class",
        "[a-]": "Unclosed character range",
    }

    for path_pat, (expected_pat, expected_simple) in pass_tests.items():
        actual_pat, actual_simple = _create_pattern_part(path_pat)
        assert actual_pat == expected_pat.replace("/", os.path.sep)
        assert actual_simple == expected_simple

    for path_pat, value_err_str in value_error_tests.items():
        try:
            _create_pattern_part(path_pat)
            assert False, "expected ValueError thrown"
        except ValueError as exc:
            assert str(exc) == value_err_str


@pytest.mark.unit
def test_create_pattern():
    """Test _create_pattern functionality"""
    pass_tests = {
        ("a/b/*.c", True): dict(
            yes=["/a", "/a/b", "/a/b/x.c", "/a/b/y.c/d"],
            no=["/a/b/x.d", "/a/x.c", "/b", "/a/b/x.cd", "/ab", "/a/b/cc"],
        ),
        ("a/b/*.c", False): dict(
            yes=["/a/b/x.c", "/a/b/y.c/d"],
            no=["/a", "/a/b", "/a/b/x.d", "/a/x.c", "/b", "/a/b/x.cd", "/ab"],
        ),
        ("a/**/b", True): dict(
            yes=["/a/b", "/a/c/b", "/a/c/d/b", "/a/c/d/b/e"],
            no=["/a", "/b", "/a/c"],
        ),
        ("a/*/b", True): dict(
            yes=["/a/c/b", "/a/c/b/e"],
            no=["/a", "/b", "/a/c", "/a/b", "/a/c/d/b"],
        ),
        ("arr[[]0].txt", False): dict(
            yes=["/arr[0].txt"],
            no=["/arr[[]0].txt", "/arr0.txt", "/arr[.txt", "/arr].txt", "/arr0].txt"],
        ),
    }

    for (path_pat, match_prefix), expected in pass_tests.items():
        pat = re.compile(
            _create_pattern(path_pat.replace("/", os.path.sep), match_prefix)
        )
        for yes_pat in expected["yes"]:
            assert pat.search(
                yes_pat.replace("/", os.path.sep)
            ), f"{path_pat},{match_prefix} should match {yes_pat}"
        for no_pat in expected["no"]:
            assert not pat.search(
                no_pat.replace("/", os.path.sep)
            ), f"{path_pat},{match_prefix} unexpectedly matched {no_pat}"

    pytest.raises(ValueError, _create_pattern, "a/[/]", True)


@pytest.mark.io
@set_umask(0)
def test_write_context():
    """Test writing a files context and hashing"""
    with tempfile.TemporaryDirectory() as tmpdir:
        make_test(
            tmpdir,
            0o777,
            {
                "subdir": (
                    0o777,
                    {
                        "bar.txt": (0o600, b"wow\n"),
                        "bar.c": (0o600, b"stuff\n"),
                        "baz.c": (
                            0o751,
                            {
                                "deepfile": (0o752, b"deepdata\n"),
                                "oth": (0o752, b"othdata\n"),
                            },
                        ),
                    },
                ),
                "data.c": (0o731, b"nice\n"),
            },
        )

        ctx = BuildContext(tmpdir, 0o022, [])
        assert (
            ctx.full_hash
            == "595b212d34c969979978e87128f79f645088057007894a5457e89b5455ecd64a"
        )
        assert len(ctx.symbolic_hash) == 64

        iob = io.BytesIO()
        ctx.write_context(iob, extra_files={"./abc": (0o642, b"abcdata")})
        iob.seek(0)
        with tarfile.open(fileobj=iob) as tf:
            assert tf.getnames() == [
                ".",
                "./data.c",
                "./subdir",
                "./subdir/bar.c",
                "./subdir/bar.txt",
                "./subdir/baz.c",
                "./subdir/baz.c/deepfile",
                "./subdir/baz.c/oth",
                "./abc",
            ]

            # Check that file exists and had its metadata updated appropriately
            ti = tf.getmember("./data.c")
            assert ti.isreg()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o755
            with tf.extractfile(ti) as tfile:
                assert tfile.read() == b"nice\n"

            ti = tf.getmember("./abc")
            assert ti.isreg()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o644
            with tf.extractfile(ti) as tfile:
                assert tfile.read() == b"abcdata"

            # Check the same for a directory

            # Check the same for a directory
            ti = tf.getmember("./subdir/baz.c")
            assert ti.isdir()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o755

        # Try again with None as the umask
        ctx = BuildContext(tmpdir, None, [])
        assert (
            ctx.full_hash
            == "eefa5a6389c48985a4cd4a4eeb7b59d653ac29502e87d9525c9d6a533498c76c"
        )
        assert len(ctx.symbolic_hash) == 64

        iob = io.BytesIO()
        ctx.write_context(iob)
        iob.seek(0)
        with tarfile.open(fileobj=iob) as tf:
            assert tf.getnames() == [
                ".",
                "./data.c",
                "./subdir",
                "./subdir/bar.c",
                "./subdir/bar.txt",
                "./subdir/baz.c",
                "./subdir/baz.c/deepfile",
                "./subdir/baz.c/oth",
            ]

            # Check that file exists and had its metadata updated appropriately
            ti = tf.getmember("./data.c")
            assert ti.isreg()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o731
            with tf.extractfile(ti) as tfile:
                assert tfile.read() == b"nice\n"

            # Check the same for a directory
            ti = tf.getmember("./subdir/baz.c")
            assert ti.isdir()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o751

        # Try one more time with ignore patterns
        ctx = BuildContext(tmpdir, 0o022, ["**/*.c", "!subdir/baz.c/deepfile"])
        assert (
            ctx.full_hash
            == "558ca8796a779d4b71742e7a53e8bd03891765c75328bf1b47ebda77666458f3"
        )
        assert len(ctx.symbolic_hash) == 64

        iob = io.BytesIO()
        ctx.write_context(iob)
        iob.seek(0)
        with tarfile.open(fileobj=iob) as tf:
            assert tf.getnames() == [
                ".",
                "./subdir",
                "./subdir/bar.txt",
                "./subdir/baz.c",
                "./subdir/baz.c/deepfile",
            ]

            # Check that file exists and had its metadata updated appropriately
            ti = tf.getmember("./subdir/bar.txt")
            assert ti.isreg()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o644
            with tf.extractfile(ti) as tfile:
                assert tfile.read() == b"wow\n"

            # Check the same for a directory
            ti = tf.getmember("./subdir/baz.c")
            assert ti.isdir()
            assert ti.uid == 0 and ti.gid == 0
            assert ti.mode == 0o755


@pytest.mark.io
def test_null_context():
    """Test using a null context"""
    ctx = BuildContext(None, 0o022, [])

    assert (
        ctx.full_hash
        == "c4a048099cc72f0af94df1e9220c21551505d8b66ef533766e06bd897d19aea4"
    )

    iob = io.BytesIO()
    ctx.write_context(iob)
    iob.seek(0)
    with tarfile.open(fileobj=iob) as tf:
        assert tf.getnames() == ["."]

        ti = tf.getmember(".")
        assert ti.isdir()
        assert ti.mode == 0o755

    iob = io.BytesIO()
    ctx.write_context(
        iob, extra_files={"./abc": (0o742, b"abcdata"), "./def": (0o600, b"defdata")}
    )
    iob.seek(0)
    with tarfile.open(fileobj=iob) as tf:
        assert tf.getnames() == [".", "./abc", "./def"]

        ti = tf.getmember(".")
        assert ti.isdir()
        assert ti.uid == 0 and ti.gid == 0
        assert ti.mode == 0o755

        ti = tf.getmember("./abc")
        assert ti.isreg()
        assert ti.uid == 0 and ti.gid == 0
        assert ti.mode == 0o755
        with tf.extractfile(ti) as tfile:
            assert tfile.read() == b"abcdata"

        ti = tf.getmember("./def")
        assert ti.isreg()
        assert ti.uid == 0 and ti.gid == 0
        assert ti.mode == 0o644
        with tf.extractfile(ti) as tfile:
            assert tfile.read() == b"defdata"



import collections
from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Tuple

from .graph import hash_graph, visit_graph
from .images import (
    BaseImage,
    ContextImage,
    CopyCommandImage,
    ImageDefinition,
    MultiPlatformImage,
    ScratchImage,
    SourceImage,
    StageData,
)


@dataclass(eq=False)
class BuildOperation:
    """
    Dataclass describing one build work unit. Each BuildOperation roughly
    corresponds to an invocation of the underlying image builder. In some
    cases it may instead just simplify to retagging existing images.

    Each operation in the chain of `image`, `image.parent`,
    `image.parent.parent`, ... up to but not including `root` will
    be part of this build unit. If `inline_context` is not None then
    it should be used as the build context.
    """

    #: The resulting image of this build operation
    image: ImageDefinition
    #: The parent image of this build operation
    root: ImageDefinition
    #: The platform to run the build operation against. An empty platform
    #: means to use the default platform (or the platform doesn't matter e.g.
    #: for context build operations).
    platform: str = ""
    #: The inline context to pass to the build, if any. A context can
    #: only be inlined if its needed nowhere else.
    inline_context: Optional[ContextImage] = None
    #: All stages associated with the resulting image
    stages: Tuple[StageData, ...] = ()
    #: All dependent build operations
    dependencies: Tuple["BuildOperation", ...] = ()


class BuildPlanner:
    """
    Class responsible for group consecutive images together that can be
    grouped and generating a dependency graph on those grouped build
    operatons.
    """

    def plan(self, stages: Iterable[StageData]) -> List[BuildOperation]:
        """
        Plan converts the rendered stage data information into a concrete
        build plan. The plan is returned as a list of build operations
        topologically sorted such that a build operation appears in the
        list after all of its dependencies.

        Arguments:
            stages: An iterable of what stages should be included in
                    the bulid plan. This can be a subset of the rendered
                    images; any required dependant stages will automatically
                    be built but their tags will not be set if not listed here.
        """
        stage_data = [
            stage
            for stage in stages
            if stage.config.image_names or stage.config.push_names
        ]
        stage_images = [stage.image for stage in stage_data]
        hash_mapping = hash_graph(stage_images)

        reverse_deps = collections.defaultdict(set)
        canonical_image: Dict[str, ImageDefinition] = {}

        def canonicalize(image: ImageDefinition) -> ImageDefinition:
            canon_image = canonical_image.setdefault(hash_mapping[image], image)
            if canon_image is not image:
                canon_image.merge_into(image)
            return canon_image

        def mark_deps(image: ImageDefinition) -> None:
            for idx, dep in enumerate(image.get_dependencies()):
                reverse_deps[dep].add(
                    (
                        idx == 0 and not isinstance(image, MultiPlatformImage),
                        image,
                    )
                )

        # Normalize all images with the same hash into the same object.
        # At the same time create a reverse dependency graph on those
        # objects.
        stage_images = visit_graph(
            stage_images,
            canonicalize,
            visit_func_post=mark_deps,
        )

        stages_by_image = collections.defaultdict(list)
        for stage, stage_image in zip(stage_data, stage_images):
            stages_by_image[stage_image].append(stage)

        build_ops: Dict[ImageDefinition, BuildOperation] = {}
        build_op_ctx_dependants = collections.defaultdict(set)
        build_op_other_dependants = collections.defaultdict(set)

        def create_op(image: ImageDefinition):
            """
            Creates all the build operations
            """
            dependants = reverse_deps.get(image, set())
            stages = tuple(stages_by_image.get(image, []))

            if (
                not stages
                and not isinstance(image, MultiPlatformImage)
                and len(dependants) == 1
                and next(iter(dependants))[0]
            ):
                # Mid-build operation image, do nothing.
                return

            # Generate build op, walking back the root as far as we can.
            root = image
            platform = ""
            build_op_ctx_deps = set()
            build_op_other_deps = set()
            while root not in build_ops:
                if isinstance(root, CopyCommandImage):
                    # Specially mark context dependencies to support inlining.
                    build_op_ctx_deps.add(build_ops[root.context])
                    root = root.parent
                    continue

                deps = root.get_dependencies()

                # Handle multi platform image nodes
                if isinstance(root, MultiPlatformImage):
                    assert image is root
                    for dep in deps:
                        build_op_other_deps.add(build_ops[dep])
                    break

                # Other images we handle generically.
                if not deps:
                    platform = getattr(root, "platform", "")
                    break
                for dep in deps[1:]:
                    build_op_other_deps.add(build_ops[dep])
                root = deps[0]
            else:
                root_dep = build_ops[root]
                platform = root_dep.platform
                build_op_other_deps.add(root_dep)

            build_op = BuildOperation(
                image=image,
                root=root,
                platform=platform,
                stages=tuple(stages_by_image.get(image, [])),
                dependencies=tuple(build_op_ctx_deps | build_op_other_deps),
            )
            for build_op_dep in build_op_ctx_deps:
                build_op_ctx_dependants[build_op_dep].add(build_op)
            for build_op_dep in build_op_other_deps:
                build_op_other_dependants[build_op_dep].add(build_op)
            build_ops[image] = build_op

        visit_graph(stage_images, lambda img: img, visit_func_post=create_op)

        removed_build_ops = set()
        for image, build_op in build_ops.items():
            if not build_op.stages and isinstance(
                image, (BaseImage, SourceImage, ScratchImage)
            ):
                # base/source images that do not correspond to a stage do not
                # represent any work so we skip them.
                removed_build_ops.add(build_op)
            elif isinstance(image, ContextImage):
                # Remove and inline contexts that are only used in one place if
                # their one use does not already have an inline.
                if (
                    build_op not in build_op_other_dependants
                    and len(build_op_ctx_dependants[build_op]) == 1
                ):
                    (dependant_build_op,) = build_op_ctx_dependants[build_op]
                    if dependant_build_op.inline_context is None:
                        dependant_build_op.inline_context = image
                        removed_build_ops.add(build_op)

        # Prune out any removed build operations from the dependency list.
        for build_op in build_ops.values():
            build_op.dependencies = tuple(
                dep_build_op
                for dep_build_op in build_op.dependencies
                if dep_build_op not in removed_build_ops
            )

        # Return the result as a list, don't need the final image as a key anymore.
        return [
            build_op
            for build_op in build_ops.values()
            if build_op not in removed_build_ops
        ]



import asyncio
import contextlib
import functools
import inspect
import logging
from contextvars import ContextVar

LOGGER = logging.getLogger(__name__)


class ScopedTaskExitStack(contextlib.AsyncExitStack):
    """
    Small extension to AsyncExitStack that can handle cancelling incomplete tasks
    cretaed during the lifetime of the stack.
    """

    def create_scoped_task(
        self,
        coro_or_task,
        *,
        propagate_exception=False,
        propagate_cancel=False,
    ):
        """
        Wrap `coro_or_task` in a task if needed and return it. Simultaneously
        add an exit callback to the exit stack to cancel and join the task
        if it has not completed when the stack exits.

        Subclasses of Exception raised by these tasks will be ignored as well as
        CancelledError. Any other errors will be propagated if there is no existing
        exception.
        """
        if inspect.iscoroutine(coro_or_task):
            task = asyncio.create_task(coro_or_task)
        else:
            task = coro_or_task

        async def cancel_task(exc_typ, exc_val, exc_tb):
            # pylint: disable=unused-argument
            if not task.done():
                task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                if propagate_cancel and exc_typ is None:
                    raise
                LOGGER.debug("Supressing cancellation error for async task")
            except BaseException as exc:  # pylint: disable=broad-except
                if propagate_exception:
                    if exc_typ is None:
                        raise
                    if exc is not exc_val:
                        LOGGER.exception(
                            "Unhandled exception in scoped async task dropped"
                        )

        self.push_async_exit(cancel_task)
        return task


_EXIT_STACK = ContextVar("exit_stack")  # type: ignore


def async_exit_context(func):
    """
    Decorator that creates a new ScopedTaskExitStack and makes it accessible
    in a context variable accessed through `get_exit_stack`. This exit stack will
    begin and end with the lifetime of the wrapped function.
    """

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        async with ScopedTaskExitStack() as stack:
            reset_token = _EXIT_STACK.set(stack)
            try:
                return await func(*args, **kwargs)
            finally:
                _EXIT_STACK.reset(reset_token)

    return wrapper


def get_exit_stack() -> ScopedTaskExitStack:
    """
    Returns the active exit stack for the current context. If no exit stack exists in
    the current context this will raise a LookupError.
    """
    return _EXIT_STACK.get()



import contextlib
import os
import shutil
import tempfile
from argparse import Namespace
from typing import Optional

import pytest

from tplbuild.cmd.base_build import BaseBuildUtility
from tplbuild.cmd.build import BuildUtility
from tplbuild.config import UserConfig
from tplbuild.tplbuild import TplBuild

BUILD_DATA_PATH = os.path.join(os.path.dirname(__file__), "builds")
TEST_CLIENT_TYPE = os.getenv("TEST_CLIENT_TYPE", "docker")
TEST_BASE_IMAGE_NAME = os.getenv("TEST_BASE_IMAGE_NAME", "localhost:8000/base")


@contextlib.asynccontextmanager
async def setup_build_test(name: str, user_config: Optional[UserConfig] = None):
    """Create the TplBuild object with default settings for the named build."""
    user_config = user_config or UserConfig(client_type=TEST_CLIENT_TYPE)
    with tempfile.TemporaryDirectory() as temp_dir:
        dst_dir = os.path.join(temp_dir, "data")
        shutil.copytree(
            os.path.join(BUILD_DATA_PATH, name),
            dst_dir,
            symlinks=True,
        )
        async with TplBuild.from_path(dst_dir, user_config=user_config) as tplbld:
            tplbld.config.base_image_repo = TEST_BASE_IMAGE_NAME
            yield dst_dir, tplbld


@pytest.mark.build
async def test_smartcopy():
    """Test that the smartcopy build works as expected."""
    params = dict(
        image=[],
        profile=[],
        platform=[],
        set_args=[],
        update_salt=False,
        update_sources=False,
        debug=False,
    )
    async with setup_build_test("smartcopy") as (base_dir, tplbld):
        result = await BaseBuildUtility().main(
            Namespace(**params, check=False),
            tplbld,
        )
        assert result == 0

        # Change file not used by base image, ensure check passes
        with open(os.path.join(base_dir, "someotherfile.txt"), "wb") as fdata:
            fdata.write(b"something something\n")
        result = await BaseBuildUtility().main(
            Namespace(**params, check=True),
            tplbld,
        )
        assert result == 0

        # Change file used by base image, ensure check fails
        with open(os.path.join(base_dir, "abc"), "wb") as fdata:
            fdata.write(b"abcdata\n")
        result = await BaseBuildUtility().main(
            Namespace(**params, check=True),
            tplbld,
        )
        assert result != 0

        # Restore file data used by base image, ensure check passes
        with open(os.path.join(base_dir, "abc"), "wb") as fdata:
            fdata.write(b"abc\n")
        result = await BaseBuildUtility().main(
            Namespace(**params, check=True),
            tplbld,
        )
        assert result == 0

        # Restore file data used by base image, ensure check passes
        with open(os.path.join(base_dir, "abc"), "wb") as fdata:
            fdata.write(b"abc\n")
        result = await BuildUtility().main(
            Namespace(**params),
            tplbld,
        )
        assert result == 0


@pytest.mark.build
async def test_multifile():
    """Test that the multifile build works as expected."""
    params = dict(image=[], profile=[], platform=[], set_args=[], debug=False)
    async with setup_build_test("multifile") as (_, tplbld):
        result = await BuildUtility().main(
            Namespace(**params),
            tplbld,
        )
        assert result == 0


@pytest.mark.build
async def test_tree():
    """Test that branching build stages work as expected."""
    params = dict(
        image=[],
        profile=[],
        platform=[],
        set_args=[],
        update_salt=False,
        update_sources=False,
        debug=False,
    )
    async with setup_build_test("tree") as (base_dir, tplbld):
        result = await BaseBuildUtility().main(
            Namespace(**params, check=False),
            tplbld,
        )
        assert result == 0

        with open(os.path.join(base_dir, "abc"), "wb") as fdata:
            fdata.write(b"abc\n")
        result = await BuildUtility().main(
            Namespace(**params),
            tplbld,
        )
        assert result == 0



import argparse
import dataclasses
from typing import Dict

from tplbuild.cmd.common import debug_build_operations
from tplbuild.cmd.utility import CliUtility
from tplbuild.exceptions import TplBuildException
from tplbuild.images import MultiPlatformImage, StageData
from tplbuild.tplbuild import TplBuild
from tplbuild.utils import compute_extra_vars


class PublishUtility(CliUtility):
    """CLI utility entrypoint for building and publishing top-level images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "Build and publish top-level images to repositories"
        parser.add_argument(
            "image",
            nargs="*",
            help="Images to build. Use 'stage_name=target_name' to "
            "override the default push name for stage_name or "
            "'stage_name=' to push the image as its stage name",
        )
        parser.add_argument(
            "--profile",
            required=False,
            default=None,
            help="Profile to build. Defaults to default profile.",
        )
        parser.add_argument(
            "--platform",
            action="append",
            help="Platform to build images for. Can be given multiple times. "
            "Defaults to all configured platforms.",
        )
        parser.add_argument(
            "--set",
            dest="set_args",
            action="append",
            type=lambda val: (False, val),
            help="Set a custom variable for the build in the format '--set a.b=c'",
        )
        parser.add_argument(
            "--set-json",
            dest="set_args",
            action="append",
            type=lambda val: (True, val),
            help="Like --set except the value will be decoded as a JSON payload."
            " e.g. '--set-json a.b=[1,\"xyz\",null]'",
        )
        parser.add_argument(
            "--debug",
            default=False,
            const=True,
            action="store_const",
            help="Only print rendered Dockerfiles instead of building images",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        profile = args.profile or tplbld.config.default_profile
        extra_vars = compute_extra_vars(args.set_args or [])
        platforms = args.platform or tplbld.config.platforms

        # Render all build stages
        multi_stage_mapping: Dict[str, StageData] = {}
        for platform in platforms:
            stage_mapping = await tplbld.render(
                profile=profile,
                platform=platform,
                extra_vars=extra_vars,
            )
            for stage_name, stage_data in stage_mapping.items():
                stage_data.config.image_names = []
                multi_stage = multi_stage_mapping.setdefault(stage_name, stage_data)

                if multi_stage is stage_data:
                    multi_stage.image = MultiPlatformImage(
                        stage_descs={
                            dataclasses.replace(desc, platform="*")
                            for desc in getattr(stage_data.image, "stage_descs", ())
                        },
                        images={platform: stage_data.image},
                    )
                    continue

                if multi_stage.config.push_names != stage_data.config.push_names:
                    raise TplBuildException(
                        "Push names must match for all platforms for stage {repr(stage_name)}"
                    )
                assert isinstance(multi_stage.image, MultiPlatformImage)
                multi_stage.image.images[platform] = stage_data.image

        # Simplify any MultiPlatformImages that only have one platform.
        for stage_data in multi_stage_mapping.values():
            assert isinstance(stage_data.image, MultiPlatformImage)
            if len(stage_data.image.images) == 1:
                stage_data.image = next(iter(stage_data.image.images.values()))

        # Figure out what images to build, override push_names where requested.
        images_to_build = set()
        for image_arg in args.image:
            image_parts = image_arg.split("=", maxsplit=1)
            images_to_build.add(image_parts[0])
            if image_parts[0] not in multi_stage_mapping:
                raise TplBuildException(f"Unknown build stage {repr(image_parts[0])}")
            if len(image_parts) > 1:
                multi_stage_mapping[image_parts[0]].config.push_names = [
                    image_parts[1] or image_parts[0],
                ]

        # Only explicitly build stages that have push_names associated with them.
        # Anything else that is needed will be included implicitly in the build graph.
        stages_to_build = [
            stage
            for stage_name, stage in multi_stage_mapping.items()
            if stage.config.push_names
            and (not images_to_build or stage_name in images_to_build)
        ]

        # Resolve the locked source image manifest content address from cached
        # build data.
        await tplbld.resolve_source_images(stages_to_build)

        # Resolve BaseImage nodes' content_hash so that their prebuilt image
        # can be referenced correctly.
        await tplbld.resolve_base_images(stages_to_build, dereference=False)

        # Create a plan of build operations to execute the requested build.
        build_ops = tplbld.plan(stages_to_build)

        # Execute the build operations.
        if args.debug:
            debug_build_operations(tplbld, build_ops)
        else:
            await tplbld.build(build_ops)

        return 0



import asyncio
import random
import time

import pytest

from tplbuild.hashing import HASHER
from tplbuild.sync_to_async_pipe import SyncToAsyncPipe


def asyncio_run(func):
    """Utility decorator to run an event loop around a coroutine"""

    def invoke(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return invoke


@pytest.mark.unit
@asyncio_run
async def test_write_bulk():
    """Test passing through the pipe using writes of differnet sizes"""
    TEST_CASES = (
        (1, 4, 4),
        (10, 20, 20),
        (1024, 128, 128),
        (2**10, 2**9, 2**9),
        (2**10, 2**4, 2**14),
        (2**10, 2**14, 2**4),
    )
    SLEEP_OFFSETS = (
        (0, 0),
        (0.1, 0),
        (0, 0.1),
    )
    rng = random.Random(555)

    for sync_sleep, async_sleep in SLEEP_OFFSETS:
        for pipe_buf_size, max_chunk_size, chunks in TEST_CASES:
            pipe = SyncToAsyncPipe(buf_max=pipe_buf_size)

            def write_sync(rng, pipe, max_chunk_size, chunks, sync_sleep):
                """Write some random data to the pipe and return its hex digest"""
                time.sleep(sync_sleep)

                hsh = HASHER()
                for _ in range(chunks):
                    data = bytes(
                        rng.randrange(256)
                        for _ in range(1 + rng.randrange(max_chunk_size))
                    )
                    hsh.update(data)
                    pipe.write(data)
                pipe.close()
                return hsh.hexdigest()

            write_task = asyncio.get_running_loop().run_in_executor(
                None, write_sync, rng, pipe, max_chunk_size, chunks, sync_sleep
            )

            await asyncio.sleep(async_sleep)

            hsh = HASHER()
            while chunk := await pipe.read():
                hsh.update(chunk)

            assert hsh.hexdigest() == await write_task



import asyncio
import logging
import os
import uuid
from asyncio.subprocess import DEVNULL, PIPE
from dataclasses import dataclass
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
)

import jinja2
from aioregistry import (
    Descriptor,
    ManifestListV2S2,
    RegistryException,
    RegistryManifestRef,
    parse_image_name,
)

from .arch import split_platform
from .config import ClientCommand
from .context import BuildContext
from .exceptions import TplBuildException
from .exit_context import async_exit_context, get_exit_stack
from .images import (
    BaseImage,
    CommandImage,
    ContextImage,
    CopyCommandImage,
    ImageDefinition,
    MultiPlatformImage,
    ScratchImage,
    SourceImage,
)
from .output import OutputStream
from .plan import BuildOperation
from .sync_to_async_pipe import SyncToAsyncPipe
from .tplbuild import TplBuild

LOGGER = logging.getLogger(__name__)


@async_exit_context
async def _create_subprocess(
    cmd: ClientCommand,
    jinja_env: jinja2.Environment,
    params: Dict[str, Any],
    *,
    capture_output: bool = False,
    output_stream: Optional[OutputStream] = None,
    input_data: Optional[AsyncIterable[bytes]] = None,
) -> bytes:
    """
    Create a subprocess and process its streams.
    """
    env = dict(os.environ)
    render_args, render_env = cmd.render(jinja_env, params)
    env.update(render_env)
    proc = await asyncio.create_subprocess_exec(
        *render_args,
        stdout=PIPE,
        stderr=PIPE,
        stdin=DEVNULL if input_data is None else PIPE,
        env=env,
    )
    assert proc.stdout is not None and proc.stderr is not None

    async def copy_lines(
        src: asyncio.StreamReader,
        *,
        output_arr: Optional[List[bytes]] = None,
        err: bool = False,
    ) -> None:
        """
        Copy lines of output from src to dst. It's assumed that dst is a non
        blocking output stream.
        """
        while not src.at_eof():
            line = await src.readline()
            if not line:
                continue
            if output_arr is not None:
                output_arr.append(line)
            if output_stream is not None:
                await output_stream.write(line, err=err)

    async def copy_input_data():
        """
        Copy data from input_data into proc.stdin.
        """
        # TODO(msg): Maybe need to disable SIGPIPE/handle write fails here?
        try:
            async for data in input_data:
                proc.stdin.write(data)
                await proc.stdin.drain()
            proc.stdin.close()
            await proc.stdin.wait_closed()
        except (BrokenPipeError, ConnectionResetError):
            LOGGER.warning("process exited before finished writing input")

    output_arr: List[bytes] = []
    coros: List[Awaitable] = [
        copy_lines(proc.stdout, output_arr=output_arr if capture_output else None),
        copy_lines(proc.stderr, err=True),
    ]
    if input_data is not None:
        coros.append(copy_input_data())

    exit_stack = get_exit_stack()
    coros.append(proc.wait())
    await asyncio.gather(*(exit_stack.create_scoped_task(coro) for coro in coros))

    if proc.returncode:
        raise TplBuildException(f"Client command failed {render_args}")

    return b"".join(output_arr)


def _construct_title(data, *, seps=":", depth=0):
    """
    Construct image titles from trie structure.
    """
    if not data:
        return ""
    children = sorted(
        (key, _construct_title(val, seps=seps, depth=depth + 1))
        for key, val in data.items()
    )
    sep = seps[depth] if depth < len(seps) else seps[-1]
    if len(children) == 1:
        return sep.join(children[0])
    if len(set(child[1] for child in children)) == 1:
        return f"{{{','.join(child[0] for child in children)}}}{sep}{children[0][1]}"
    return ",".join(f"{{{child[0]}{sep}{child[1]}}}" for child in children)


def _compute_titles(build_ops: List[BuildOperation]) -> List[str]:
    all_profiles: Set[str] = set()
    all_platforms: Set[str] = set()

    for build_op in build_ops:
        descs = getattr(build_op.image, "stage_descs", ())
        all_profiles.update(desc.profile for desc in descs)
        all_platforms.update(desc.platform for desc in descs)

    titles = []
    for build_op in build_ops:
        descs = getattr(build_op.image, "stage_descs", ())

        hierarchy: dict = {}
        for desc in descs:
            parts = [desc.name]
            if len(all_profiles) > 1:
                parts.append(desc.profile)
            if len(all_platforms) > 1 and not isinstance(
                build_op.image, MultiPlatformImage
            ):
                osname, arch, var = split_platform(desc.platform)
                if var:
                    arch = f"{arch}/{var}"
                parts.append(osname)
                parts.append(arch)

            data = hierarchy
            for part in parts:
                data = data.setdefault(part, {})

        if not hierarchy:
            titles.append("intermediate")
            continue

        titles.append(
            _construct_title(
                hierarchy,
                seps="::/" if len(all_profiles) > 1 else ":/",
            )[:-1]
        )

    return titles


@dataclass(eq=False)
class RenderedBuildOperation:
    """
    Dataclass describing a rendered build operation.
    """

    #: Dockerfile data to pass to underlying builder.
    dockerfile: str
    #: Tags to set locally. If a tag is maps to True it should be pushed.
    tags: Dict[str, bool]
    #: Tag to initially build the image as. This may be a transient tag if
    #: there was no requested tag associated with this build operation.
    primary_tag: str
    #: User friendly title for the build operation.
    build_title: str
    #: Set to True if the build operation does not actually create a new image.
    #: This is the case for example when a base image is created solely from
    #: a source image.
    build_empty: bool


class BuildExecutor:
    """
    Utility class that acts as an interface between tplbuild and the client
    build commands.
    """

    def __init__(
        self,
        tplbld: TplBuild,
    ) -> None:
        self.tplbld = tplbld
        self.transient_prefix = "tplbuild"

        user_config = tplbld.user_config
        self.client_config = user_config.client
        self.sem_build_jobs = asyncio.BoundedSemaphore(user_config.build_jobs)
        self.sem_push_jobs = asyncio.BoundedSemaphore(user_config.push_jobs)
        self.sem_tag_jobs = asyncio.BoundedSemaphore(user_config.tag_jobs)
        self.build_retry = user_config.build_retry
        self.push_retry = user_config.push_retry

    @async_exit_context
    async def build(
        self,
        build_ops: List[BuildOperation],
        *,
        complete_callback: Optional[
            Callable[[BuildOperation, str], Awaitable[None]]
        ] = None,
    ) -> None:
        """
        Build each of the passed build ops and tag/push all images.

        Arguments:
            build_ops: The list of build operations to be completed. These build
                operations should be topologically sorted (every build operation
                should be listed after all of its dependencies).
            complete_callback: If present this callback will be invoked for each
                build operation as `await complete_callback(build_op, primary_tag)`.
                Note for multi platform images the 'primary_tag' will just be the
                first push tag.
        """
        transient_images: List[str] = []
        remote_pull_coros: Dict[str, Awaitable] = {}

        rendered_ops = self.render_build_ops(build_ops)
        image_tag_map: Dict[ImageDefinition, str] = {
            build_op.image: rendered_op.primary_tag
            for build_op, rendered_op in zip(build_ops, rendered_ops)
            if not rendered_op.build_empty
        }

        async def _build_single(
            build_op: BuildOperation, rendered_op: RenderedBuildOperation
        ):
            # Wait for dependencies to finish
            for dep in build_op.dependencies:
                await build_done_events[dep].wait()

            tags = rendered_op.tags
            primary_tag = rendered_op.primary_tag
            build_title = rendered_op.build_title

            if isinstance(build_op.image, MultiPlatformImage):
                await self._build_multi_platform(
                    build_op.image, tags, image_tag_map, build_title
                )
                build_done_events[build_op].set()
            else:
                if isinstance(build_op.image, ContextImage):
                    await self._build_context(primary_tag, build_op.image, build_title)
                else:
                    # Pull base images, source images
                    remote_deps, local_deps = self._get_build_deps(
                        build_op, image_tag_map
                    )
                    if self.client_config.pull is not None:
                        for remote_ref, remote_name in remote_deps.items():
                            if remote_ref not in remote_pull_coros:
                                remote_pull_coros[remote_ref] = asyncio.create_task(
                                    self.pull_image(remote_ref, remote_name)
                                )
                            await remote_pull_coros[remote_ref]

                    await self._build_work(
                        primary_tag,
                        build_op,
                        rendered_op.dockerfile,
                        local_deps,
                        build_title,
                    )
                build_done_events[build_op].set()

                if not tags:
                    transient_images.append(primary_tag)

                for tag, push in tags.items():
                    if tag != primary_tag:
                        await self.tag_image(primary_tag, tag)
                    if push:
                        await self.push_image(tag, build_title)

            if complete_callback:
                await complete_callback(build_op, primary_tag)

        stack = get_exit_stack()
        build_tasks = {
            build_op: stack.create_scoped_task(_build_single(build_op, rendered_op))
            for build_op, rendered_op in zip(build_ops, rendered_ops)
        }
        build_done_events = {build_op: asyncio.Event() for build_op in build_ops}
        try:
            done, _ = await asyncio.wait(
                build_tasks.values(),
                return_when=asyncio.FIRST_EXCEPTION,
            )
            # Propagate any exceptions that occurred
            for task in done:
                await task
        finally:
            await asyncio.gather(
                *(self.untag_image(image) for image in transient_images)
            )

    async def _build_multi_platform(
        self,
        image: MultiPlatformImage,
        tags: Dict[str, bool],
        image_tag_map: Dict[ImageDefinition, str],
        title: str,
    ) -> None:
        """
        Push a multi-architecture image with the given tags. All tags
        must be push tags for this kind of node.
        """
        if not all(push for push in tags.values()):
            raise TplBuildException("Multi platform images only support push tags")

        async def push_sub_image(
            image_ref: RegistryManifestRef,
            platform: str,
            sub_image: ImageDefinition,
        ) -> Descriptor:
            sub_image_ref = image_ref.copy(
                update=dict(ref=f"{image_ref.ref}-{platform.replace('/', '-')}")
            )
            sub_image_tag = image_tag_map[sub_image]
            await self.tag_image(sub_image_tag, str(sub_image_ref))
            await self.push_image(str(sub_image_ref), f"{title}:{platform}")
            try:
                desc = await self.tplbld.registry_client.ref_lookup(sub_image_ref)
            except RegistryException as exc:
                raise TplBuildException("Failed to look up image digest") from exc
            if desc is None:
                raise TplBuildException("Could not look up pushed image on registry")
            return desc

        for tag in tags:
            image_ref = parse_image_name(tag)
            sub_descriptors = await asyncio.gather(
                *(
                    push_sub_image(image_ref, platform, sub_image)
                    for platform, sub_image in image.images.items()
                )
            )
            sub_manifest_items = []
            for platform, sub_descriptor in zip(image.images, sub_descriptors):
                image_os, architecture, variant = split_platform(platform)
                sub_manifest_items.append(
                    dict(
                        platform=dict(
                            architecture=architecture,
                            os=image_os,
                            variant=variant,
                        ),
                        **sub_descriptor.dict(by_alias=True, exclude_unset=True),
                    )
                )

            manifest = ManifestListV2S2(
                schemaVersion=2,
                mediaType="application/vnd.docker.distribution.manifest.list.v2+json",
                manifests=sub_manifest_items,
            )
            await self.tplbld.registry_client.manifest_write(image_ref, manifest)
            async with self.tplbld.output_streamer.start_stream(title) as output_stream:
                await output_stream.write(
                    f"Wrote multi architecture platform {image_ref}".encode("utf-8")
                )

    async def _build_context(self, tag: str, image: ContextImage, title: str) -> None:
        """
        Perform a build operation where the image is an ImageContext.
        """
        await self.client_build(
            tag,
            image.platform,
            b"FROM scratch\nCOPY . /\n",
            title,
            context=image.context,
        )

    def _get_build_deps(
        self,
        build_op: BuildOperation,
        image_tag_map: Dict[ImageDefinition, str],
    ) -> Tuple[Dict[str, str], Set[str]]:
        """
        Return a list of remotely stored images this build depends on.
        """
        remote_deps = {}
        local_deps = set()
        img = build_op.image

        def _title_image(img):
            if isinstance(img, BaseImage):
                return f"{img.stage}:{img.profile}:{img.platform}"
            assert isinstance(img, SourceImage)
            return f"{img.repo}:{img.tag}:{img.platform}"

        while img is not build_op.root:
            if (
                isinstance(img, CopyCommandImage)
                and img.context is not build_op.inline_context
            ):
                image_name = self._name_image(img.context, image_tag_map)
                if isinstance(img.context, (BaseImage, SourceImage)):
                    remote_deps[image_name] = _title_image(img.context)
                else:
                    local_deps.add(image_name)

            img = img.parent  # type: ignore

        image_name = self._name_image(img, image_tag_map)
        if isinstance(img, (BaseImage, SourceImage)):
            remote_deps[image_name] = _title_image(img)
        else:
            local_deps.add(image_name)

        return remote_deps, local_deps

    def render_build_ops(
        self,
        build_ops: List[BuildOperation],
    ) -> List[RenderedBuildOperation]:
        """
        Render Dockerfiles for each build operation.
        """
        result: List[RenderedBuildOperation] = []
        image_tag_map: Dict[ImageDefinition, str] = {}

        for build_op, build_title in zip(build_ops, _compute_titles(build_ops)):
            # Construct mapping of all tags to a bool indicating if the
            # tag should be pushed. The dict is ordered in the same order
            # as the stages with the tags from each stage keeping the same
            # relative order with push tags after tags.
            tags: Dict[str, bool] = {}
            for stage in build_op.stages:
                for tag in stage.config.image_names or []:
                    tags.setdefault(tag, False)
                for tag in stage.config.push_names or []:
                    tags[tag] = True

            primary_tag = ""
            if isinstance(build_op.image, MultiPlatformImage):
                result.append(
                    RenderedBuildOperation(
                        "# Multi-arch image", tags, primary_tag, build_title, True
                    )
                )
                continue

            if tags:
                primary_tag = next(iter(tags))
            else:
                primary_tag = f"{self.transient_prefix}-{uuid.uuid4()}"

            if isinstance(build_op.image, ContextImage):
                result.append(
                    RenderedBuildOperation(
                        "# Shared context image", tags, primary_tag, build_title, False
                    )
                )
                image_tag_map[build_op.image] = primary_tag
                continue

            lines = []

            img = build_op.image
            while img is not build_op.root:
                if isinstance(img, CommandImage):
                    lines.append(f"{img.command} {img.args}")
                    img = img.parent
                elif isinstance(img, CopyCommandImage):
                    if img.context is build_op.inline_context:
                        lines.append(f"COPY {img.command}")
                    else:
                        lines.append(
                            f"COPY --from={ self._name_image(img.context, image_tag_map) } {img.command}"
                        )
                    img = img.parent
                else:
                    raise AssertionError("Unexpected image type in build operation")

            build_empty = not lines
            lines.append(f"FROM { self._name_image(img, image_tag_map) }")
            if syntax := self.tplbld.config.dockerfile_syntax:
                lines.append(f"# syntax={syntax}")

            result.append(
                RenderedBuildOperation(
                    "\n".join(reversed(lines)),
                    tags,
                    primary_tag,
                    build_title,
                    build_empty,
                )
            )
            if not build_empty:
                image_tag_map[build_op.image] = primary_tag

        return result

    async def _build_work(
        self,
        tag: str,
        build_op: BuildOperation,
        dockerfile: str,
        local_deps: Set[str],
        title: str,
    ) -> None:
        """
        Perform a build operation as a series of Dockerfile commands.
        """
        await self.client_build(
            tag,
            build_op.platform,
            dockerfile.encode("utf-8"),
            title,
            context=build_op.inline_context.context
            if build_op.inline_context
            else None,
            dependencies=local_deps,
        )

    def _name_image(
        self, image: ImageDefinition, image_tag_map: Dict[ImageDefinition, str]
    ) -> str:
        """
        Construct the name of an image from its ImageDefinition. `image` should always be
        either an ExternalImage or the resulting image of a previously calculated
        bulid operation.
        """
        tag = image_tag_map.get(image)
        if tag is not None:
            return tag
        if isinstance(image, SourceImage):
            assert image.digest is not None
            return f"{image.repo}@{image.digest}"
        if isinstance(image, BaseImage):
            return self.tplbld.get_base_image_name(image, use_digest=True)
        if isinstance(image, ScratchImage):
            return "scratch"
        raise AssertionError("unexpected image type")

    async def client_build(
        self,
        image: str,
        platform: str,
        dockerfile_data: bytes,
        title: str,
        *,
        context: Optional[BuildContext] = None,
        dependencies: Optional[Set[str]] = None,
    ) -> None:
        """Wrapper that executes the client command to start a build"""

        async with self.sem_build_jobs:
            if context is None:
                context = BuildContext(None, None, [])

            async with self.tplbld.output_streamer.start_stream(title) as output_stream:
                for attempt in range(self.build_retry + 1):

                    def sync_write_context(pipe: SyncToAsyncPipe):
                        assert context is not None
                        context.write_context(
                            pipe,  # type: ignore
                            extra_files={"Dockerfile": (0o444, dockerfile_data)},
                        )
                        pipe.close()

                    async def pipe_reader(pipe: SyncToAsyncPipe):
                        try:
                            while data := await pipe.read():
                                yield data
                        finally:
                            pipe.close()

                    pipe = SyncToAsyncPipe()
                    try:
                        await asyncio.gather(
                            asyncio.get_running_loop().run_in_executor(
                                None, sync_write_context, pipe
                            ),
                            _create_subprocess(
                                self.client_config.build,
                                self.tplbld.jinja_env,
                                dict(
                                    image=image,
                                    platform=platform,
                                    dependencies=dependencies or set(),
                                ),
                                output_stream=output_stream,
                                input_data=pipe_reader(pipe),
                            ),
                        )
                        break
                    except TplBuildException:
                        await output_stream.write(
                            f"Build failed on attempt {attempt + 1}/{self.build_retry + 1}".encode(
                                "utf-8"
                            )
                        )
                        if attempt == self.build_retry:
                            raise

    async def tag_image(self, source_image: str, target_image: str) -> None:
        """Wrapper that executes the client tag command"""
        async with self.sem_tag_jobs:
            await _create_subprocess(
                self.client_config.tag,
                self.tplbld.jinja_env,
                dict(source_image=source_image, target_image=target_image),
            )

    async def untag_image(self, image: str) -> None:
        """Wrapper that executes the client untag command"""
        async with self.sem_tag_jobs:
            await _create_subprocess(
                self.client_config.untag,
                self.tplbld.jinja_env,
                dict(image=image),
            )

    async def pull_image(self, image: str, title: str) -> None:
        """Wrapper that executes the client pull command"""
        assert self.client_config.pull is not None
        async with self.sem_push_jobs:
            async with self.tplbld.output_streamer.start_stream(title) as output_stream:
                for attempt in range(self.push_retry + 1):
                    try:
                        await _create_subprocess(
                            self.client_config.pull,
                            self.tplbld.jinja_env,
                            dict(image=image),
                            output_stream=output_stream,
                        )
                    except TplBuildException:
                        await output_stream.write(
                            f"Pull failed on attempt {attempt + 1}/{self.push_retry + 1}".encode(
                                "utf-8"
                            )
                        )
                        if attempt == self.push_retry:
                            raise

    async def push_image(self, image: str, title: str) -> None:
        """Wrapper that executes the client push command"""
        async with self.sem_push_jobs:
            async with self.tplbld.output_streamer.start_stream(title) as output_stream:
                for attempt in range(self.push_retry + 1):
                    try:
                        await _create_subprocess(
                            self.client_config.push,
                            self.tplbld.jinja_env,
                            dict(image=image),
                            output_stream=output_stream,
                        )
                    except TplBuildException:
                        await output_stream.write(
                            f"Pull failed on attempt {attempt + 1}/{self.push_retry + 1}".encode(
                                "utf-8"
                            )
                        )
                        if attempt == self.push_retry:
                            raise

    async def platform(self) -> str:
        """
        Returns the platform of the build daemon or an empty string if it
        cannot be determined. No normalization of the returned platform is done.
        """
        if self.client_config.platform is None:
            return ""

        output = await _create_subprocess(
            self.client_config.platform,
            self.tplbld.jinja_env,
            {},
        )
        try:
            return output.decode("utf-8").strip()
        except UnicodeDecodeError as exc:
            raise TplBuildException("Failed to decode executor platform") from exc
